from matbench.constants import CLF_KEY, REG_KEY
from matbench.data_ops import load

from copy import deepcopy
import pprint

from monty.serialization import loadfn


mbv01_metadata = loadfn("/Users/ardunn/alex/lbl/projects/common_env/dev_codes/matbench/matbench/matbench_v0.1_dataset_metadata.json")
new_metadata = deepcopy(mbv01_metadata)

for ds, info in mbv01_metadata.items():
    print(f"starting {ds}")
    all_data = load(ds)
    target_data = all_data[info["target"]]
    if info["task_type"] == REG_KEY:
        new_metadata[ds]["mad"] = target_data.mad()
    elif info["task_type"] == CLF_KEY:
        new_metadata[ds]["frac_true"] = target_data.mean()
    else:
        raise ValueError


pprint.pprint(new_metadata)



import unittest

import numpy as np
from pymatgen.core import Structure

from matbench.constants import CLF_KEY, REG_KEY
from matbench.data_ops import (
    homogenize_clf_array,
    load,
    mean_absolute_percentage_error,
    score_array,
)
from matbench.metadata import mbv01_metadata
from matbench.tests.util import FULL_TEST


class TestDataOps(unittest.TestCase):
    def test_downloads_mbv01(self):

        if FULL_TEST:
            metadata = mbv01_metadata
        else:
            restricted_set = [
                "matbench_dielectric",
                "matbench_steels",
                "matbench_glass",
            ]
            metadata = {
                k: v for k, v in mbv01_metadata.items() if k in restricted_set
            }

        for dataset, metadata in metadata.items():
            df = load(dataset)

            self.assertEqual(df.shape[0], metadata["n_samples"])
            self.assertIn(metadata["target"], df.columns)

            input_type = metadata["input_type"]
            test_type = Structure if input_type == "structure" else str
            self.assertTrue(
                all(
                    [
                        isinstance(input, test_type)
                        for input in df[metadata["input_type"]]
                    ]
                )
            )

            problem_type = metadata["task_type"]
            test_types = (
                [np.bool_, bool]
                if problem_type == CLF_KEY
                else [np.float_, np.float32, np.float64, float]
            )
            self.assertIn(df[metadata["target"]].dtypes, test_types)

    def test_score_array(self):
        # test for regression
        true = [1, 2, 3, 4]
        test = [1, 3, 3, 4]
        ans = score_array(true, test, task_type=REG_KEY)
        true_ans = {"mae": 0.25, "rmse": 0.5, "mape": 0.125, "max_error": 1}
        self.assertDictEqual(ans, true_ans)

        # test for classification
        true = [True, False]
        test = [True, True]
        ans = score_array(true, test, task_type=CLF_KEY)
        true_ans = {
            "accuracy": 0.5,
            "balanced_accuracy": 0.5,
            "f1": 0.6666666666666666,
            "rocauc": 0.5,
        }
        self.assertDictEqual(ans, true_ans)

        # test for probability clf
        true = [True, False]
        test = [0.7, 0.65]
        ans = score_array(true, test, task_type=CLF_KEY)
        self.assertDictEqual(ans, true_ans)

    def test_mean_absolute_percentage_error(self):

        true = [1, 100, 1000, 0, 0.00000001]
        test = [0.9, 81, 1010, 1, 0.0028933]

        # make sure the small or zero values are masked
        mape = mean_absolute_percentage_error(true, test, threshold=1e-6)
        mape_masked = mean_absolute_percentage_error(true[:4], test[:4])
        self.assertAlmostEqual(mape, 0.09999999999999999)
        self.assertAlmostEqual(mape, mape_masked)

    def test_homogenize_clf_array(self):

        bools = [True, False, True, True]
        floats = [1.0, 0.3, 0.5001, 0.9]

        probs = homogenize_clf_array(bools, to_probs=True)
        self.assertAlmostEqual(probs[0], 1.0, places=5)
        self.assertAlmostEqual(probs[1], 0.0, places=5)
        self.assertAlmostEqual(probs[2], 1.0, places=5)
        self.assertAlmostEqual(probs[3], 1.0, places=5)

        labels = homogenize_clf_array(floats, to_labels=True, thresh=0.5)
        self.assertTrue(labels[0])
        self.assertFalse(labels[1])
        self.assertTrue(labels[2])
        self.assertTrue(labels[3])

        labels2 = homogenize_clf_array(floats, to_labels=True, thresh=0.91)
        self.assertTrue(labels2[0])
        self.assertFalse(labels2[1])
        self.assertFalse(labels2[2])
        self.assertFalse(labels2[3])



import os
import random

import numpy as np

from matbench.constants import CLF_KEY, REG_KEY

MB_TEST_RANDOM_SEED = 1001
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
FULL_TEST = os.environ.get("MB_FULL_TESTS", False)


def model_random(training_outputs, test_inputs, response_type):
    r = random.Random(MB_TEST_RANDOM_SEED)

    r = np.random.RandomState(MB_TEST_RANDOM_SEED)

    l = len(test_inputs)

    if response_type == CLF_KEY:
        return r.choice([True, False], size=l)

    # Regression: simply sample from random distribution bounded by max
    # and min training samples
    if response_type == REG_KEY:
        pred = r.uniform(max(training_outputs), min(training_outputs), size=l)
        return pred



"""
for use with the coverage job in this repo's github workflows
"""

import os

from monty.serialization import loadfn


fname = os.path.join(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir), "coverage.json")
coverage = loadfn(fname)

total_percentage = coverage["totals"]["percent_covered"]

if total_percentage < 75.0:
    raise ValueError(f"Coverage of {total_percentage} is less than threshold, 75%.")


"""
Core class for benchmarking.
"""

import datetime
import logging
import pprint
import traceback

from monty.json import MSONable

from matbench.constants import (
    CLF_KEY,
    COMPOSITION_KEY,
    MBV01_KEY,
    REG_KEY,
    STRUCTURE_KEY,
    VERSION,
)
from matbench.metadata import mbv01_metadata
from matbench.task import MatbenchTask
from matbench.util import (
    MSONable2File,
    RecursiveDotDict,
    hash_dictionary,
    immutify_dictionary,
    initialize_logger,
)

logger = initialize_logger(logger_name="matbench", level=logging.INFO)


class MatbenchBenchmark(MSONable, MSONable2File):
    """The core class for benchmarking with Matbench.

    MatbenchBenchmark is capable of benchmarking and validating arbitrary
    materials science benchmarks. It is a container class for sets of
    MatbenchTasks, objects which provide predetermined sets of
    training/validation and testing data for any algorithm to benchmark
    with. MatbenchBenchmark can also give summaries of entire complex
    benchmarks, including access to individual score statistics for
    each metric.

    MatbenchBenchmark can run any benchmark as long as it has a corresponding
    benchmark name key. Matbench v0.1 ("matbench_v0.1") is the only benchmark
    currently configured for use with MatbenchBenchmark.

    MatbenchBenchmark is capable of running benchmark subsets; for example,
    only 3 of the 13 available Matbench v0.1 problems.

    See the documentation for more details.

    Attributes:
        benchmark_name (str): The benchmark name, defaults to the original
            Matbench v0.1 "matbench_v0.1". Should have an associated
            validation file in order for the MatbenchTasks to work
            correctly.
        metadata (dict): The corresponding metadata file for this benchmark,
            which defines the basic configuration for each task. See
            matbench_v0.1_validation for an example. Each dataset
            has the same required keys in order to work correctly.
        user_metadata (dict): Any metadata about the algorithm or benchmark
            that the user wants to keep as part of the benchmark file.
        tasks_map ({str: MatbenchTask}): A mapping of task name to the
            corresponding MatbenchTask object.

        <<task_names>> (MatbenchTask): Access any task obj via
            MatbenchTask.<<task_name>>. For example:

            mb = MatbenchBenchmark()
            mb.matbench_dielectric

            <<MatbenchTask object>>
    """

    # For serialization
    _VERSION_KEY = "version"
    _BENCHMARK_KEY = "benchmark_name"
    _USER_METADATA_KEY = "user_metadata"
    _TASKS_KEY = "tasks"
    _DATESTAMP_KEY = "datestamp"
    _DATESTAMP_FMT = "%Y.%m.%d %H:%M.%S"
    _HASH_KEY = "hash"

    # For class usage
    ALL_KEY = "all"

    def __init__(self, benchmark=MBV01_KEY, autoload=False, subset=None):
        """

        Args:
            benchmark (str): The name of the benchmark. Only supported benchmark
                currently is "matbench_v0.1", though more will be added in the
                future.
            autoload (bool): If True, automatically load the dataset into memory
                For a full benchmark, this can take some time. If False, you'll
                need to load each task with .load before you can access the raw
                data.
            subset ([str]): A list of task names to use as a subset of a full
                benchmark. Only the named tasks will be contained in the class.
                Must correspond to the metadata file defined by the benchmark
                name.
        """

        if benchmark == MBV01_KEY:
            self.benchmark_name = MBV01_KEY
            self.metadata = mbv01_metadata
        else:
            raise ValueError(
                f"Only '{MBV01_KEY}' available. No other benchmarks defined!"
            )

        if subset:
            not_datasets = [k for k in subset if k not in self.metadata]
            if not_datasets:
                raise KeyError(
                    f"Some tasks in {subset} are not benchmark="
                    f"'{self.benchmark_name}' datasets! Remove {not_datasets}."
                )
            else:
                available_tasks = subset
        else:
            available_tasks = self.metadata.keys()

        self.user_metadata = {}
        self.tasks_map = RecursiveDotDict()

        for ds in available_tasks:
            self.tasks_map[ds] = MatbenchTask(
                ds, autoload=autoload, benchmark=self.benchmark_name
            )

        logger.info(
            f"Initialized benchmark '{benchmark}' "
            f"with {len(available_tasks)} tasks: \n"
            f"{pprint.pformat(list(available_tasks))}"
        )

    def __getattr__(self, item):
        """
        Enable MatbenchBenchmark.task_name behavior.

        Args:
            item (str): The name of the attr.

        Returns:
            The attr, if not in the metadata defined by the benchmark
            If the attr is a task name, returns that MatBenchTask object.

        """
        if item in self.metadata:
            return self.tasks_map[item]
        else:
            return self.__getattribute__(item)

    @classmethod
    def from_preset(cls, benchmark, preset_name, autoload=False):
        """
        The following presets are defined for each benchmark:

        benchmark: 'matbench_v0.1':

            - preset: 'structure' - Only structure problems
            - preset: 'composition' - Only composition problems
            - preset: 'regression' - Only regression problems
            - preset: 'classification' - Only classification problems
            - preset: 'all' - All problems in matbench v0.1

        Args:
            benchmark (str): Name of the benchmark set you'd like to use. The
                only supported benchmark set currently is "matbench_v0.1"
            preset_name (str): The name of the preset
            autoload (bool): If true, automatically loads all the datasets
                upon instantiation. Be warned; this can take a while.

        Returns:
            (MatbenchBenchmark object)

        """
        if benchmark == MBV01_KEY:
            if preset_name == STRUCTURE_KEY:
                available_tasks = [
                    k
                    for k, v in mbv01_metadata.items()
                    if v.input_type == STRUCTURE_KEY
                ]
            elif preset_name == COMPOSITION_KEY:
                available_tasks = [
                    k
                    for k, v in mbv01_metadata.items()
                    if v.input_type == COMPOSITION_KEY
                ]
            elif preset_name == REG_KEY:
                available_tasks = [
                    k for k, v in mbv01_metadata.items() if v.task_type == REG_KEY
                ]
            elif preset_name == CLF_KEY:
                available_tasks = [
                    k for k, v in mbv01_metadata.items() if v.task_type == CLF_KEY
                ]
            elif preset_name == cls.ALL_KEY:
                available_tasks = [k for k, v in mbv01_metadata.items()]
            else:
                valid_keys = [
                    STRUCTURE_KEY,
                    COMPOSITION_KEY,
                    CLF_KEY,
                    REG_KEY,
                    cls.ALL_KEY,
                ]
                raise ValueError(
                    f"Preset name '{preset_name}' not recognized for "
                    f"benchmark '{MBV01_KEY}'! Select from "
                    f"{valid_keys}"
                )
        else:
            raise ValueError(
                f"Only '{MBV01_KEY}' available. No other benchmarks defined!"
            )

        return cls(benchmark=benchmark, autoload=autoload, subset=available_tasks)

    @classmethod
    def from_dict(cls, d):
        """Create a MatbenchBenchmark object from a dictionary.

        Args:
            d (dict):

        Returns:
            (MatbenchBenchmark)

        """
        required_keys = [
            "@module",
            "@class",
            cls._VERSION_KEY,
            cls._BENCHMARK_KEY,
            cls._TASKS_KEY,
            cls._USER_METADATA_KEY,
            cls._DATESTAMP_KEY,
            cls._HASH_KEY,
        ]

        missing_keys = []
        for k in required_keys:
            if k not in d:
                missing_keys.append(k)

        extra_keys = []
        for k in d:
            if k not in required_keys:
                extra_keys.append(k)

        if missing_keys and not extra_keys:
            raise ValueError(
                f"Required keys {missing_keys} for {cls.__class__.__name__} "
                f"not found!"
            )
        elif not missing_keys and extra_keys:
            raise ValueError(
                f"Extra keys {extra_keys} for {cls.__class__.__name__} " f"present!"
            )
        elif missing_keys and extra_keys:
            raise ValueError(
                f"Missing required keys {missing_keys} and extra keys "
                f"{extra_keys} present!"
            )

        # Check all tasks to make sure their benchmark name is matching in the
        # benchmark and in the tasks
        not_matching_bench = []
        for t_dict in d[cls._TASKS_KEY].values():
            if t_dict[MatbenchTask._BENCHMARK_KEY] != d[cls._BENCHMARK_KEY]:
                not_matching_bench.append(t_dict[MatbenchTask._DATASET_KEY])
        if not_matching_bench:
            raise ValueError(
                f"Tasks {not_matching_bench} do not have a benchmark name "
                f"matching the benchmark ({d[cls._BENCHMARK_KEY]})!"
            )

        # Ensure the hash is matching, i.e., the data was not modified after
        # matbench got done with it
        m_from_dict = d.pop(cls._HASH_KEY)
        m = hash_dictionary(d)
        if m != m_from_dict:
            raise ValueError(
                f"Hash of dictionary does not match it's reported value! {m} "
                f"!= {m_from_dict} . Was the data modified after saving?)"
            )

        # Check to see if any tasks have task names not matching their key
        # names in the benchmark
        not_matching_tasks = []
        for task_name, task_info in d[cls._TASKS_KEY].items():
            key_as_per_task = task_info[MatbenchTask._DATASET_KEY]
            if task_name != key_as_per_task:
                not_matching_tasks.append((task_name, key_as_per_task))
        if not_matching_tasks:
            raise ValueError(
                f"Task names in benchmark and task names in tasks not "
                f"matching: {not_matching_tasks}"
            )

        # Warn if versions are not matching
        if d[cls._VERSION_KEY] != VERSION:
            logger.warning(
                f"Warning! Versions not matching: "
                f"(data file has version {d[cls._VERSION_KEY]}, "
                f"this package is {VERSION})."
            )

        return cls._from_args(
            benchmark_name=d[cls._BENCHMARK_KEY],
            tasks_dict=d[cls._TASKS_KEY],
            user_metadata=d[cls._USER_METADATA_KEY],
        )

    @classmethod
    def _from_args(cls, benchmark_name, tasks_dict, user_metadata):
        """Create a MatbenchBenchmark object from arguments

        Args:
            benchmark_name (str): name of the benchmark
            tasks_dict (dict): formatted dict of task data
            user_metadata (dict): freeform user metadata

        Returns:
            (MatbenchBenchmark)
        """
        subset = list(tasks_dict.keys())
        obj = cls(benchmark=benchmark_name, autoload=False, subset=subset)
        obj.tasks_map = RecursiveDotDict(
            {
                t_name: MatbenchTask.from_dict(t_dict)
                for t_name, t_dict in tasks_dict.items()
            }
        )

        logger.warning(
            "To add new data to this benchmark, the "
            "benchmark must be loaded with .load(). Alternatively, "
            "load individual tasks with MatbenchTask.load()."
        )

        # MatbenchTask automatically validates files during its from_dict
        obj.user_metadata = user_metadata

        logger.debug(f"Successfully converted dict/args to '{cls.__name__}'.")

        return obj

    def _determine_completeness(self, completeness_type):
        """Determine the completeness of this benchmark.

        Completeness means the tasks are included (but not
        necessarily recorded yet) in the benchmark.

        Supported completeness types are:
        - "all": All tasks are included
        - "composition": All composition tasks are included
        - "structure": All structure tasks are included
        - "regression": All regression problems
        - "classification": All classification problems

        Args:
            completeness_type (str): One of the above completeness
                types.

        Returns:
            (bool) True if this benchmark object is complete
                with respect to the completeness type.

        """
        if completeness_type == self.ALL_KEY:
            required_tasks = list(self.metadata.keys())
        elif completeness_type in (COMPOSITION_KEY, STRUCTURE_KEY):
            required_tasks = [
                k
                for k, v in self.metadata.items()
                if v.input_type == completeness_type
            ]
        elif completeness_type in (REG_KEY, CLF_KEY):
            required_tasks = [
                k
                for k, v in self.metadata.items()
                if v.task_type == completeness_type
            ]
        else:
            allowed_completeness_types = [
                self.ALL_KEY,
                COMPOSITION_KEY,
                STRUCTURE_KEY,
                REG_KEY,
                CLF_KEY,
            ]
            raise ValueError(
                "Only supported completeness types are "
                f"{allowed_completeness_types}"
            )

        for task in required_tasks:
            if task not in self.tasks_map:
                return False
        else:
            return True

    def as_dict(self):
        """Overridden from MSONable.as_dict, get dict repr of this obj

        Returns:
            (dict)

        """
        tasksd = {mbt.dataset_name: mbt.as_dict() for mbt in self.tasks}
        tasksd_jsonable = immutify_dictionary(tasksd)

        d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            self._VERSION_KEY: VERSION,
            self._TASKS_KEY: tasksd_jsonable,
            self._USER_METADATA_KEY: self.user_metadata,
            self._BENCHMARK_KEY: self.benchmark_name,
            self._DATESTAMP_KEY: datetime.datetime.utcnow().strftime(
                self._DATESTAMP_FMT
            ),
        }

        # to obtain a hash for this benchmark, immutify the dictionary
        # and then stringify it
        d[self._HASH_KEY] = hash_dictionary(d)
        logger.debug(
            f"Successfully converted {self.__class__.__name__} to dictionary."
        )
        return d

    def get_info(self):
        """Log info about the benchmark to the respective logging handlers.

        Returns:
            None
        """
        logger.info(self.info)

    def add_metadata(self, metadata):
        """Add freeform information about this run to the object
        (and subsequent json), accessible thru the
        'user_metadata' attr.


        All keys must be strings.

        All values must be either:
            a. a numpy ndarray
            b. python native types, such as bools, floats, ints, strs
            c. a pandas series
            d. a list/tuple of python native types (bools, floats, ints)

            OR

            e. A dictionary where all keys are strs and all values
               are one of a, b, c, d, or e (recursive).

        Args:
            metadata (dict)

        Returns:
            None
        """
        # Use logging here so bad metadata addition does not
        # ruin an entire run...
        if not isinstance(metadata, dict):
            logger.critical(
                f"User metadata must be reducible to dict format, "
                f"not type({type(metadata)})"
            )
            logger.info("User metadata not added.")

        else:
            if self.user_metadata:
                logger.warning("User metadata already exists! Overwriting...")

            self.user_metadata = immutify_dictionary(metadata)
            logger.info("User metadata added successfully!")

    def load(self):
        """Load all tasks in this benchmark.
        Returns:
            None
        """
        for t in self.tasks:
            t.load()

    def validate(self):
        """Run validation on each task in this benchmark.

        Returns:
            ({str: str}): dict of errors, if they exist

        """
        errors = {}
        for t, t_obj in self.tasks_map.items():
            try:
                t_obj.validate()
            except BaseException:
                errors[t] = traceback.format_exc()
        return errors

    @property
    def tasks(self):
        """Return the tasks as a list.

        Returns:
            ([MatbenchTask]): A list of matbench tasks in this benchmark
        """
        return self.tasks_map.values()

    @property
    def scores(self):
        """Get all score metrics for all tasks as a dictionary.

        Returns:
            (RecursiveDotDict): A nested dictionary-like object of scores
                for each task.

        """
        return RecursiveDotDict({t.dataset_name: t.scores for t in self.tasks})

    @property
    def info(self):
        """Get a formatted string of info about this benchmark and its current
        state.

        Returns:
            (str)

        """

        complete = self.is_complete
        recorded = self.is_recorded
        valid = self.is_valid

        s = ""
        s += (
            f"\nMatbench package {VERSION} running benchmark "
            f"'{self.benchmark_name}'"
        )
        s += f"\n\tis complete: {complete}"
        s += f"\n\tis recorded: {recorded}"
        s += f"\n\tis valid: {valid}"

        if not recorded:
            s += (
                "\n\n Benchmark is not fully recorded; limited information " "shown."
            )
        if not valid:
            s += "\n\n Benchmark is not valid; limited information shown."

        if not valid or not recorded:
            s += "\n\nTasks:"
            for t in self.tasks_map.values():
                s += f"\n\t- '{t.dataset_name}: recorded={t.all_folds_recorded}"

        if valid and recorded:
            s += "\n\nResults:"
            for t in self.tasks:

                if t.metadata.task_type == REG_KEY:
                    score_text = (
                        f"MAE mean: " f"{self.scores[t.dataset_name].mae.mean}"
                    )
                else:
                    score_text = (
                        f"ROCAUC mean: " f"{self.scores[t.dataset_name].rocauc.mean}"
                    )
                s += f"\n\t- '{t.dataset_name}' {score_text}"

        return s

    @property
    def is_complete(self):
        """Determine if all available tasks are included in this benchmark.

        For matbench v0.1, this means all 13 tasks are in the benchmark.

        Returns:
            (bool)

        """
        return self._determine_completeness(completeness_type=self.ALL_KEY)

    @property
    def is_composition_complete(self):
        """Determine if all composition tasks for this benchmark are included

        Returns:
            (bool)
        """
        return self._determine_completeness(completeness_type=COMPOSITION_KEY)

    @property
    def is_structure_complete(self):
        """Determine if all structure tasks for this benchmark are included

        Returns:
            (bool)
        """
        return self._determine_completeness(completeness_type=STRUCTURE_KEY)

    @property
    def is_regression_complete(self):
        """Determine if all regression tasks for this benchmark are included

        Returns:
            (bool)
        """
        return self._determine_completeness(completeness_type=REG_KEY)

    @property
    def is_classification_complete(self):
        """Determine if all classification tasks for this benchmark are included

        Returns:
            (bool)
        """
        return self._determine_completeness(completeness_type=CLF_KEY)

    @property
    def is_recorded(self):
        """All tasks in this benchmark (whether or not it includes all tasks in
        the benchmark set) are recorded.

        Returns:
            (bool): True if all tasks (even if only a subset of all matbench)
            for this benchmark are recorded.

        """
        return all([t.all_folds_recorded for t in self.tasks_map.values()])

    @property
    def is_valid(self):
        """Checks all tasks are recorded and valid, as per each task's
        validation procedure.

        Can take some time, especially if the tasks are not already
        loaded into memory.

        Returns:
            (bool): True if all tasks are valid
        """
        errors = self.validate()
        if errors:
            formatted_errors = pprint.pformat(errors)
            logger.critical(
                f"Benchmark has errors! " f"Errors:\n {formatted_errors}"
            )
            return False
        else:
            return True



import json
import os
import glob
import logging
import pprint
from operator import gt, lt

import tqdm
from monty.serialization import loadfn
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import numpy as np

from matbench.task import MatbenchTask
from matbench.bench import MatbenchBenchmark
from matbench.constants import MBV01_KEY, CLF_KEY, REG_KEY
from matbench.metadata import mbv01_metadata

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DOCS_DIR = os.path.join(THIS_DIR, "../docs_src")
STATIC_DOCS_DIR = os.path.join(DOCS_DIR, "static")
BENCHMARKS_DIR = os.path.join(THIS_DIR, "../benchmarks")
FULL_DATA_DIR = os.path.join(DOCS_DIR, "Full Benchmark Data")
PER_TASK_DIR = os.path.join(DOCS_DIR, "Leaderboards Per-Task")
PER_TASK_DIR_PREFIX = "Leaderboards%20Per-Task/"
FULL_DATA_DIR_PREFIX = "Full%20Benchmark%20Data/"
METADATA_DIR = os.path.join(DOCS_DIR, "Benchmark Info")
METADATA_DIR_PREFIX = "Benchmark%20Info/"
SNIPPETS_DIR = os.path.join(THIS_DIR, "doc_snippets")
SCALED_ERRORS_FILENAME = "scaled_errors_{bmark_name}.html"


MP_WEBSITE_STATICS = os.path.join(STATIC_DOCS_DIR, "mp_srcs")
SCALED_ERRORS_PATH = os.path.join(STATIC_DOCS_DIR, SCALED_ERRORS_FILENAME)
SCALED_ERRORS_JSON_PATH = SCALED_ERRORS_PATH.replace(".html", ".json")


def generate_scaled_errors_graph(gp_graph_data_by_bmark):
    """
    Generate the scaled errors graph


    Args:
        gp_graph_data_by_bmark:

    Returns:

    """
    for bmark_name, tasks_data in gp_graph_data_by_bmark.items():

            if bmark_name == MBV01_KEY:

                symbols = {
                    "matbench_steels": "σᵧ",
                    "matbench_jdft2d": "Eˣ",
                    "matbench_phonons": "ωᵐᵃˣ",
                    "matbench_dielectric": "𝑛",
                    "matbench_expt_gap": "Eᵍ",
                    "matbench_expt_is_metal": "Expt. Metallicity",
                    "matbench_glass": "Metallic Glass",
                    "matbench_log_kvrh": "log₁₀Kᵛʳʰ",
                    "matbench_log_gvrh": "log₁₀Gᵛʳʰ",
                    "matbench_perovskites": "Eᶠ",
                    "matbench_mp_gap": "Eᵍ",
                    "matbench_mp_is_metal": "Metallicity",
                    "matbench_mp_e_form": "Eᶠ"
                }

                descriptors = {
                    "matbench_steels": "Steel alloys",
                    "matbench_jdft2d": "2D Materials",
                    "matbench_phonons": "Phonons",
                    "matbench_dielectric": "",
                    "matbench_expt_gap": "Experimental",
                    "matbench_expt_is_metal": "Classification",
                    "matbench_glass": "Classification",
                    "matbench_log_gvrh": "",
                    "matbench_log_kvrh": "",
                    "matbench_perovskites": "Perovskites, DFT",
                    "matbench_mp_gap": "DFT",
                    "matbench_mp_is_metal": "DFT",
                    "matbench_mp_e_form": "DFT"
                }

                metadata = mbv01_metadata
            else:
                raise ValueError(
                    f"Only {MBV01_KEY} defined as valid benchmark! '{bmark_name}' not supported.")

            # should take care of missing entries (i.e., structure only) automatically
            df = pd.DataFrame(tasks_data)

            # make scaled data for heatmap coloring
            # scale regression problems by mad/mae
            def scale_regression_problem(series, mad):
                mask = series > 0.0
                mask_iix = np.where(mask)
                series.iloc[mask_iix] = series.iloc[mask_iix] / mad
                series.loc[~mask] = np.nan
                return series

            def scale_classification_problem(series, mad):
                mask = series > 0.0
                mask_iix = np.where(mask)
                series.iloc[mask_iix] = 1 - (series.iloc[mask_iix] - 0.5) / 0.5
                # series.iloc[mask_iix] = 1 - series.iloc[mask_iix]
                series.loc[~mask] = np.nan
                return series

            scaled_df = df.copy(deep=True)
            for task in scaled_df.columns:
                if metadata[task].task_type == CLF_KEY:
                    scaler = scale_classification_problem
                else:
                    scaler = scale_regression_problem

                scaled_df[task] = scaler(scaled_df[task], metadata[task].mad)

            scaled_df = scaled_df.T
            scaled_df["n_samples"] = [metadata[task].n_samples for task in scaled_df.index]
            scaled_df["Problem"] = [f"{symbols[task]} {descriptors[task]}" for task in scaled_df.index]
            scaled_df = scaled_df.sort_values(by="n_samples")
            scaled_df.index = scaled_df["Problem"]
            scaled_df = scaled_df.drop(columns=["n_samples", "Problem"])

            best_values = scaled_df.min(axis=1)
            best_algos = scaled_df.idxmin(axis=1)

            fig = px.scatter(scaled_df, log_y=True)
            fig.update_traces(
                marker={'size': 10},
                hovertemplate="<br>".join(
                    [
                        # "Algorithm: %{text}",
                        "Problem: %{x}",
                        "Scaled Error: %{y}"
                    ]
                )
            )

            fig.update_layout(title_text="Scaled Errors",
                              title_font_size=30,
                              legend_font_size=15,
                              legend_title_font_size=15,
                              legend_title_text="Algorithm",
                              yaxis_title="Scaled MAE (regression) or <br> (1-ROCAUC)/0.5 (classification)",
                              xaxis_title="", paper_bgcolor='rgba(0,0,0,0)',
                              plot_bgcolor='rgba(0,0,0,0)',
                              font={"color": "white"})

            # add scatter for the best algorithms on scaled error
            fig.add_trace(
                go.Scatter(
                    mode="markers",
                    x=best_values.index,
                    y=best_values,
                    marker=dict(
                        color="yellow",
                        size=10,
                    ),
                    hovertemplate= \
                        'Algorithm: %{text}<br>Problem: %{x}<br>Scaled Error: %{y}<br>',
                    text=best_algos,
                    visible="legendonly",
                    name="Best algorithms"
                )
            )
            fig.update_xaxes(linecolor="grey", gridcolor="grey")
            fig.update_yaxes(linecolor="grey", gridcolor="grey")
            fig.write_html(SCALED_ERRORS_PATH.format(bmark_name=bmark_name))


            # Update layout for showing on white background on mp website
            fig.update_layout(
                title_text="",
                font={"color": "black"}
            )
            fig.write_json(SCALED_ERRORS_JSON_PATH.format(bmark_name=bmark_name))


# NOTE: MUST BE CALLED AFTER CREATING generate_scaled_errors_graph
def generate_general_purpose_leaderboard_and_plot(gp_leaderboard_data_by_bmark):
    """
    Generate both the general purpose leaderboard and scaled errors graph on the main leaderboard page.


    THE SCALED ERRORS GRAPH MUST EXIST FOR THIS FUNCTION TO COMPLETE CORRECTLY.
    Call generate_scaled_errors_graph to create this graph.

    This function takes in the doc snippet for index.md and prepends the leaderboard to it.

    So to make changes to the index.md you should change the doc snippet, not the index.md file in docs_src.


    Args:
        gp_leaderboard_data_by_bmark:

    Returns:

    """
    gp_leaderboard_txt = ""
    for bmark, gp_data in gp_leaderboard_data_by_bmark.items():

        if bmark == MBV01_KEY:
            metadata = mbv01_metadata
        else:
            raise ValueError(f"{bmark} not a valid benchmark!")

        table_data = {
            "task": [k for k in gp_data.keys()],
            "n_samples": [metadata[k].n_samples for k in gp_data.keys()],
            "algorithm": [d["algorithm"] for d in gp_data.values()],
            "completeness": [d["completeness"] for d in gp_data.values()],
            "link": [d["link"] for d in gp_data.values()],
            "score": [d["score"] for d in gp_data.values()],
            "type": [d["type"] for d in gp_data.values()]
        }

        df_src = pd.DataFrame(table_data).sort_values(by="n_samples")
        table_header = f"## Leaderboard: General Purpose Algorithms on `{bmark}`\n\n"
        table_explanation = f"Find more information about this benchmark on [the benchmark info page]({METADATA_DIR_PREFIX}{bmark}.md)\n\n"
        table = "| Task name | Samples | Algorithm | Verified MAE (unit) or ROCAUC | Notes |\n" \
                "|------------------|---------|-----------|----------------------|-------|\n"
        # create leaderboard table
        for _, row in df_src.iterrows():

            task_name = f"{row['task']}"
            task_name_link = f"[{task_name}]({PER_TASK_DIR_PREFIX}matbench_v0.1_{task_name}.md)"
            samples = format_int(row["n_samples"])
            algorithm = f"[{row['algorithm']}]({row['link']}.md)"

            task_metadata = metadata[row['task']]

            score = f"{format_float(row['score'])} ({task_metadata.unit})" if task_metadata.task_type == REG_KEY else f"{format_float(row['score'])}"
            score = f"**{score}**"

            if row["completeness"] == "structure":
                notes = "structure required"
            elif row["completeness"] == "all":
                notes = ""
            else:
                raise ValueError(f"{row['completeness']} is not a valid type of general purpose completeness!")

            table += f"| {task_name_link} | {samples} | {algorithm} | {score} | {notes} |\n"
        table += "\n\n"
        scaled_errors_plot_txt = f'\n<iframe src="static/{SCALED_ERRORS_FILENAME.format(bmark_name=bmark)}" class="is-fullwidth" height="700px" width="1000px" frameBorder="0"> </iframe>\n\n'
        gp_leaderboard_txt += table_header + table_explanation + table + scaled_errors_plot_txt

    # Load the static index from the snippets dir
    with open(os.path.join(SNIPPETS_DIR, "index.md"), encoding="utf-8") as f:
        static_txt = f.read()


    page_header = f"# Leaderboard\n\n"
    final_txt = page_header + gp_leaderboard_txt + static_txt

    with open(os.path.join(DOCS_DIR, "index.md"), "w", encoding="utf-8") as f:
        print("Writing leaderboard and plot to index.md...")
        f.write(final_txt)

    with open(os.path.join(STATIC_DOCS_DIR,  "gp_table.json"), "w", encoding="utf-8") as f:
        print("Writing static gp_table json to gp_table.json...")
        j = {"txt": gp_leaderboard_txt}
        json.dump(j, f)


def generate_per_task_leaderboards(task_leaderboard_data_by_bmark):
    """
    Generate all .md files for the per-task leaderboards.


    Args:
        task_leaderboard_data_by_bmark:

    Returns:

    """
    for bmark_name, tasks_data in task_leaderboard_data_by_bmark.items():
        for task, entries in tasks_data.items():

            task_type = entries[0]["type"]

            if task_type == REG_KEY:

                table_data = {
                    "algorithm": [],
                    "algorithm w/ link": [],
                    "mean mae": [],
                    "std mae": [],
                    "mean rmse": [],
                    "max max_error": []
                }
            elif task_type == CLF_KEY:
                table_data = {
                    "algorithm": [],
                    "algorithm w/ link": [],
                    "mean rocauc": [],
                    "std rocauc": [],
                    "mean f1": [],
                    "mean balanced_accuracy": []
                }
            else:
                raise ValueError(f"Task type {task_type} not recognized!")

            for entry in entries:
                algo_name = entry["algorithm"]
                # link must be relative, and since this is inside a dir, has to reference outside
                link = "../" + entry["link"]
                scores = entry["scores"]

                table_data["algorithm w/ link"].append(f"[{algo_name}]({link})")
                table_data["algorithm"].append(algo_name)

                if task_type == REG_KEY:
                    table_data["mean mae"].append(scores.mae.mean)
                    table_data["std mae"].append(scores.mae.std)
                    table_data["mean rmse"].append(scores.rmse.mean)
                    table_data["max max_error"].append(scores.max_error.max)
                else:
                    table_data["mean rocauc"].append(scores.rocauc.mean)
                    table_data["std rocauc"].append(scores.rocauc.std)
                    table_data["mean f1"].append(scores.f1.mean)
                    table_data["mean balanced_accuracy"].append(scores.balanced_accuracy.mean)

            df = pd.DataFrame(table_data)
            df = df.set_index("algorithm")
            sorting_column = "mean rocauc" if task_type == CLF_KEY else "mean mae"
            sorting_order = True if task_type == REG_KEY else False
            df = df.sort_values(by=sorting_column, ascending=sorting_order)

            mbt = MatbenchTask(task, autoload=False, benchmark=bmark_name)

            # header of the page
            header = f"# {bmark_name} {task}\n\n"

            subheader = f"## Individual Task Leaderboard for `{task}`\n\n"
            explanation = "_Leaderboard for an individual task. Algorithms shown here may include " \
                          "both general purpose and specialized algorithms (i.e., algorithms which " \
                          "are only valid for a subset of tasks in the benchmark._\n\n"

            info_header = "### Dataset info\n\n"

            info_body = f"##### Description\n\n{mbt.metadata.description}\n\n"
            info_body += f"Number of samples: {mbt.metadata.n_samples}\n\n"
            info_body += f"Task type: {mbt.metadata.task_type}\n\n"
            info_body += f"Input type: {mbt.metadata.input_type}\n\n"
            info_body += f"##### Dataset columns\n\n" + "".join([f"- {c}: {cd}\n" for c, cd in mbt.metadata.columns.items()]) + "\n\n"
            info_body += f"##### Dataset reference\n\n `{mbt.metadata.reference}`\n\n"

            metadata_header = "### Metadata\n\n"
            metadata = f"```\n{pprint.pformat(mbt.metadata)}\n```\n\n"

            table_header = "### Leaderboard\n\n"
            column_headers = df.columns
            table = "| " + " | ".join(column_headers) + " |\n" +  "|------" * len(column_headers) + "|\n"
            for ix, row in df.iterrows():
                table += "| "
                for th in column_headers:
                    if th == "algorithm w/ link":
                        table += f"{row[th]} | "
                    else:
                        number = format_float(row[th])
                        if th == sorting_column:
                            number = f"**{number}**"
                        table += f"{number} | "
                table += "\n"
            table += "\n"

            table = table.replace("algorithm w/ link", "algorithm")


            droppables = ["mean f1", "mean balanced_accuracy"] if task_type == CLF_KEY else ["mean rmse", "max max_error"]
            df = df.drop(columns=["algorithm w/ link"] + droppables)
            error_metric = "std rocauc" if task_type == CLF_KEY else "std mae"
            fig = px.scatter(df, error_y=error_metric, log_y=True)

            metric = f"MAE {mbt.metadata.unit}" if task_type == REG_KEY else "ROCAUC"

            target = mbt.metadata.target
            lower_or_higher = "lower" if task_type == REG_KEY else "higher"
            title_prefix = "Errors predicting" if task_type == REG_KEY else \
                "Classification ROCAUC predicting"
            title = f"{title_prefix} '{target}' ({lower_or_higher} is better)"
            fig.update_layout(title_text=title,
                              title_font_size=15,
                              showlegend=False,
                              yaxis_title=metric,
                              xaxis_title="", paper_bgcolor='rgba(0,0,0,0)',
                              plot_bgcolor='rgba(0,0,0,0)',
                              font={"color": "white"})
            fig.update_yaxes(linecolor="grey", gridcolor="grey")
            fig.update_xaxes(linecolor="rgba(0,0,0,0)", gridcolor="rgba(0,0,0,0)")

            fig_path = f"task_{bmark_name}_{task}.html"
            fig.write_html(os.path.join(STATIC_DOCS_DIR, fig_path))

            fig_reference = f'\n<iframe src="../../static/{fig_path}" class="is-fullwidth" height="700px" width="1000px" frameBorder="0"> </iframe>\n\n'

            task_leaderboard_page = header + \
                                    subheader + \
                                    explanation + \
                                    table_header + \
                                    table + \
                                    fig_reference + \
                                    info_header + \
                                    info_body + \
                                    metadata_header + \
                                    metadata
            fname = os.path.join(PER_TASK_DIR, f"{bmark_name}_{task}.md")
            with open(fname, "w", encoding="utf-8") as f:
                print(f"Creating task leaderboard page {fname}")
                f.write(task_leaderboard_page)


def organize_task_data(all_data):
    """
    Preprocessing step for organizing all benchmark data in specific formats,
    to be handed downstream to .md-generating functions.


    Args:
        all_data:

    Returns:

    """
    all_data_per_benchmark = {}

    prefix = FULL_DATA_DIR_PREFIX

    for data_packet in all_data.values():
        bmark_name = data_packet["results"].benchmark_name
        if bmark_name in all_data_per_benchmark:
            all_data_per_benchmark[bmark_name].append(data_packet)
        else:
            all_data_per_benchmark[bmark_name] = [data_packet]


    gp_leaderboard_data_by_bmark = {}
    task_leaderboards_data_by_bmark = {}
    gp_graph_data_by_bmark = {}

    for bmark_name, bmarks in all_data_per_benchmark.items():
        if bmark_name == MBV01_KEY:
            metadata = mbv01_metadata
        else:
            raise ValueError(f"No other benchmarks configured ('{bmark_name}')")

        gp_leaderboard = {t: {
            "score": None,
            "type": None,
            "link": None,
            "algorithm": None,
            "completeness": None
        } for t in metadata.keys()}

        task_leaderboards = {t: [] for t in metadata.keys()}

        gp_graph_data = {t: {} for t in metadata.keys()}

        for bmark_data in bmarks:
            mb = bmark_data["results"]
            info = bmark_data["info"]
            dir_name_short = bmark_data["dir_name_short"]

            for task in mb.tasks:
                task_name = task.dataset_name

                if task.metadata.task_type == REG_KEY:
                    score = task.scores.mae.mean

                    # Better regression tasks have lower mean mae
                    op = lt
                elif task.metadata.task_type == CLF_KEY:
                    score = task.scores.rocauc.mean

                    # Better classification tasks have higher mean rocauc
                    op = gt
                else:
                    raise ValueError


                # Include both GP, structure-required, and strcuture/comp regression algos on
                # GP leaderboard, as there are 9 structure problems or 10 regresison problems
                # across multiple dataset sizes (composition only or clf only are not included)
                if mb.is_complete or mb.is_structure_complete or mb.is_regression_complete:
                    current_best_score = gp_leaderboard[task_name]["score"]

                    gp_graph_data[task_name][info["algorithm"]] = score

                    # this task's score is better or it is the first so far
                    if current_best_score is None or op(score, current_best_score):
                        gp_leaderboard[task_name]["score"] = score
                        gp_leaderboard[task_name]["link"] = prefix + dir_name_short
                        gp_leaderboard[task_name]["algorithm"] = info["algorithm"]
                        gp_leaderboard[task_name]["type"] = task.metadata.task_type

                        if mb.is_complete or mb.is_regression_complete:
                            gp_leaderboard[task_name]["completeness"] = "all"
                        else:
                            gp_leaderboard[task_name]["completeness"] = "structure"
                    # the existing task score is best
                    else:
                        pass

                # Add it to the task-specific leaderboard, as all entries will be included
                # there
                task_leaderboards[task_name].append({
                    "scores": task.scores,
                    "link": prefix + dir_name_short + ".md",
                    "algorithm": info["algorithm"],
                    "type": task.metadata.task_type
                })

            gp_leaderboard_data_by_bmark[bmark_name] = gp_leaderboard
            task_leaderboards_data_by_bmark[bmark_name] = task_leaderboards
            gp_graph_data_by_bmark[bmark_name] = gp_graph_data

    return gp_leaderboard_data_by_bmark, task_leaderboards_data_by_bmark, gp_graph_data_by_bmark


def generate_metadata_pages(task_leaderboard_data_by_bmark):
    """
    Generate all benchmark info pages. E.g., the table for matbench_v0.1 with all the download links.

    There should be one per benchmark version - e.g., one for matbench_v0.1, one for matbench_v0.2, ...


    Args:
        task_leaderboard_data_by_bmark:

    Returns:

    """
    for bmark_name, bmark_data in task_leaderboard_data_by_bmark.items():
        metadata = MatbenchBenchmark(benchmark=bmark_name, autoload=False).metadata

        d = {}
        for task, infod in metadata.items():
            d[task] = {
                "Task name": f"`{task}`",
                "Task type": infod.task_type,
                "Target column (unit)": f"`{infod.target}` " + f"({infod.unit})" if infod.unit else f"`{infod.target}`",
                "Input type": infod.input_type,
                "Samples": infod.n_samples,
                "MAD (regression) or Fraction True (classification)": format_float(infod.mad if infod.task_type == REG_KEY else infod.frac_true),
                "Links": f"[download](https://ml.materialsproject.org/projects/{task}.json.gz), [interactive](https://ml.materialsproject.org/projects/{task})",
                "Submissions": f"{len(bmark_data[task])}"
            }

        df = pd.DataFrame(d).T.sort_values(by="Samples")

        df["Samples"] = [format_int(i) for i in df["Samples"]]
        table_header = f"# Benchmark info for `{bmark_name}`\n\n"
        table_explanation = f"The `{bmark_name}` benchmark contains {len(metadata)} tasks:\n\n"
        table = "| " + " | ".join(df.columns) + "|\n" + \
            "|-------" * len(df.columns) + "|\n"
        for _, row in df.iterrows():
            table_line = "|"
            for c in df.columns:
                table_line += f" {row[c]} |"

            table_line += "\n"
            table += table_line

        page = table_header + table_explanation + table

        path = os.path.join(METADATA_DIR, f"{bmark_name}.md")
        with open(path, "w", encoding="utf-8") as f:
            print(f"Writing benchmark info page {path}")
            f.write(page)


def generate_info_pages(all_data):
    """
    Generate all full data benchmarks pages. There should be one per submitted benchmark.

    E.g., one for /benchmarks/matbench_v0.1_algo1, one for /benchmarks/matbench_v0.1_algo2, etc.

    Args:
        all_data:

    Returns:

    """
    for bmark_name, bmark_data in tqdm.tqdm(all_data.items(), desc="DOCS: FULL DATA DOCS GENERATED"):
        info = bmark_data["info"]
        mb = bmark_data["results"]
        dir_name_short = bmark_data["dir_name_short"]

        doc_str = generate_info_page(mb, info, dir_name_short)

        doc_path = os.path.join(FULL_DATA_DIR, f"{dir_name_short}.md")
        with open(doc_path, "w", encoding="utf-8") as f:
            print(f"Writing full benchmark data page {doc_path}")
            f.write(doc_str)


def generate_info_page(mb: MatbenchBenchmark, info: dict, dir_name_short: str):
    """
    Generate a single full data benchmark page.


    Args:
        mb:
        info:
        dir_name_short:

    Returns:

    """
    is_complete = convert_bool_to_unicode_check(mb.is_complete)
    structure_complete = convert_bool_to_unicode_check(mb.is_structure_complete)
    composition_complete = convert_bool_to_unicode_check(mb.is_composition_complete)
    regression_complete = convert_bool_to_unicode_check(mb.is_regression_complete)
    classification_complete = convert_bool_to_unicode_check(mb.is_classification_complete)

    algo_name = info["algorithm"]
    algo_desc = info["algorithm_long"]
    refs = info["bibtex_refs"]
    notes = info["notes"]
    requirements = info["requirements"]

    header = f"# {mb.benchmark_name}: {algo_name}\n\n"
    url = f"https://github.com/hackingmaterials/matbench/tree/main/benchmarks/{dir_name_short}"
    desc = f"### Algorithm description: \n\n{algo_desc}\n\n#### Notes:\n{notes}\n\nRaw data download and example notebook available [on the matbench repo]({url}).\n\n"
    refs = f"### References (in bibtex format): \n\n```\n{pprint.pformat(refs)}\n```\n\n"

    user_metadata = f"### User metadata:\n\n```\n{pprint.pformat(mb.user_metadata)}\n```\n\n"

    n_tasks_available = len(mb.tasks)
    n_tasks_total = len(mb.metadata.keys())

    metadata_header = f"### Metadata:\n\n"

    metadata_table = f"| tasks recorded | {n_tasks_available}/{n_tasks_total} |\n" \
                     f"|----------------|-------------------------------------|\n" \
                     f"| complete? | {is_complete} | \n" \
                     f"| composition complete? | {composition_complete} | \n" \
                     f"| structure complete? | {structure_complete} | \n" \
                     f"| regression complete? | {regression_complete} | \n" \
                     f"| classification complete? | {classification_complete} | \n\n"

    metadata_header += metadata_table

    requirements_header = f"### Software Requirements\n\n"
    requirements_body = f"```\n{pprint.pformat(requirements)}\n```\n\n"

    all_tasks_header = f"### Task data:\n\n"
    data_txt = ""
    for task in mb.tasks:
        task_header = f"#### `{task.dataset_name}`\n\n"

        fold_data_header = f"###### Fold scores\n\n"

        # needed score order as the score order is not same between fold scores and task scores
        score_order = list(task.scores.keys())
        score_order_display = ["mape*" if s == "mape" else s for s in score_order]
        fold_table = "| fold | " + " | ".join(score_order_display) + " |\n" + \
                     "|------ " * (len(score_order) + 1) + "|\n"
        for fold_key, fold_data in task.results.items():
            fold_line = f" | {fold_key} "

            for metric_name in score_order:
                metric_val = fold_data.scores[metric_name]
                fold_line += f"| {format_float(metric_val)}"
            fold_line += " |\n"
            fold_table += fold_line
        fold_table += "\n\n"



        fold_dist_header = f"###### Fold score stats\n\n"
        dist_table = "| metric | mean | max | min | std |\n" \
                     "|--------|------|-----|-----|-----|\n"
        for metric_name, stats in task.scores.items():
            # add an asterisk next to mape since the metric is edited to not skew data on very small magnitude values
            display_name = metric_name + "*" if metric_name == "mape" else metric_name
            dist_table += f"| {display_name} | {format_float(stats.mean)} | {format_float(stats.max)} | {format_float(stats.min)} | {format_float(stats.std)} |\n"

        dist_table += "\n\n"

        params_header = "###### Fold parameters\n\n"
        params_table = "| fold | params dict|\n" \
                       "|------|------------|\n"
        for fold_key, fold_data in task.results.items():
            params_short = f"{fold_data.parameters}"[:200]
            params_short = params_short + "..." if len(params_short) == 200 else params_short
            fold_line = f"| {fold_key} | `{params_short}` |\n"
            params_table += fold_line

        params_table += "\n\n"

        task_section = task_header + fold_data_header + fold_table + fold_dist_header + dist_table + params_header + params_table + "\n\n"
        data_txt += task_section

    final_txt = header + desc + refs + user_metadata + metadata_header + requirements_header + requirements_body + all_tasks_header + data_txt
    return final_txt


def format_float(number):
    return f"{number:.4f}"


def format_int(number):
    return f'{number:,}'


def nuke_docs(check=True):
    """
    Clean the source data directory.

    Args:
        check (bool): don't actually delete anything, just output the files to-be-deleted.

    Returns:

    """
    count = 0

    index = os.path.join(DOCS_DIR, "index.md")
    if os.path.exists(index):
        if not check:
            os.remove(index)
        print(f"\tdeleting index md file'{index}'")
    count += 1

    html_statics = glob.glob(os.path.join(STATIC_DOCS_DIR, "*.html"))
    for hs in html_statics:
        if not check:
            os.remove(hs)
        print(f"\tdeleting static html '{hs}'")
    count += len(html_statics)

    json_statics = glob.glob(os.path.join(STATIC_DOCS_DIR, "*.json"))
    for js in json_statics:
        if not check:
            os.remove(js)
        print(f"\tdeleting static json '{js}'")
    count += len(json_statics)

    per_task_leaderboards = glob.glob(os.path.join(PER_TASK_DIR, "*.md"))
    for pl in per_task_leaderboards:
        if not check:
            os.remove(pl)
        print(f"\tdeleting per task leaderboard md file '{pl}'")
    count += len(per_task_leaderboards)

    full_benchmark_mds = glob.glob(os.path.join(FULL_DATA_DIR, "*.md"))
    for fb in full_benchmark_mds:
        if not check:
            os.remove(fb)
        print(f"\tdeleting full benchmark md file '{fb}'")
    count += len(full_benchmark_mds)

    benchmark_info_mds = glob.glob(os.path.join(METADATA_DIR, "*.md"))
    for bi in benchmark_info_mds:
        if "notes.md" not in bi:
            if not check:
                os.remove(bi)
            print(f"\tdeleting benchmark info page '{bi}'")
            count += 1

    print(f"\tdeleted {count} files from {DOCS_DIR}")


def convert_bool_to_unicode_check(t_or_f):
    return "✓" if t_or_f else "✗"


if __name__ == "__main__":

    logging.root.setLevel(logging.DEBUG)

    # nuke_docs(check=True)
    nuke_docs(check=False)

    all_data = {}

    # Get all benchmark data loaded into memory
    # If throws an error trying to obtain any of this data,
    # Should abort the whole docs build
    for d in os.listdir(BENCHMARKS_DIR):
        if d not in [".DS_Store", ".ipynb_checkpoints"]:
            print(d)
            d_path = os.path.join(BENCHMARKS_DIR, d)
            results_path = os.path.join(d_path, "results.json.gz")
            info_path = os.path.join(d_path, "info.json")

            # results are automatically validated, no need to validate again
            mb = MatbenchBenchmark.from_file(results_path)

            info = loadfn(info_path)

            name = info["algorithm"]
            all_data[name] = {"results": mb, "info": info, "dir_name_short": d}


    gp_leaderboard_data_by_bmark, task_leaderboards_data_by_bmark, gp_graph_data_by_bmark = organize_task_data(all_data)


    print("DOCS: ALL DATA ACQUIRED")

    generate_info_pages(all_data)

    generate_per_task_leaderboards(task_leaderboards_data_by_bmark)

    generate_metadata_pages(task_leaderboards_data_by_bmark)

    # must be called before generating the gp leaderboard
    generate_scaled_errors_graph(gp_graph_data_by_bmark)

    generate_general_purpose_leaderboard_and_plot(gp_leaderboard_data_by_bmark)





"""
Utils and quality-of-life enhancements.
"""

import datetime
import hashlib
import json
import logging
import os
import sys

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


class RecursiveDotDict(dict):
    """
    Adapted from user Curt Hagenlocher from
    https://stackoverflow.com/questions/3031219/recursively
    -access-dict-via-attributes-as-well-as-index-access
    """

    MARKER = object()

    def __init__(self, value=None):
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError("expected dict")

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, RecursiveDotDict):
            value = RecursiveDotDict(value)
        super(RecursiveDotDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        found = self.get(key, RecursiveDotDict.MARKER)
        if found is RecursiveDotDict.MARKER:
            found = RecursiveDotDict()
            super(RecursiveDotDict, self).__setitem__(key, found)
        return found

    __setattr__, __getattr__ = __setitem__, __getitem__

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)


class MSONable2File:
    """
    Adds some very basic to_file and from_file methods common to
    all msonables in this package.
    """

    def to_file(self, filename):
        with open(filename, "w") as f:
            d = self.as_dict()
            json.dump(d, f)

        logger.info(
            f"Successfully wrote {self.__class__.__name__} " f"to file '{filename}'."
        )

    @classmethod
    def from_file(cls, filename):
        with open(filename, "r") as f:
            d = json.loads(f.read())
        return cls.from_dict(d)


def initialize_logger(logger_name, log_dir=None, level=None) -> logging.Logger:
    """Initialize the default logger with stdout and file handlers.
    Args:
        logger_name (str): The package name.
        log_dir (str, None): Path to the folder where the log file will be
            written. If None, no file will be written.
        level (int): The log level. For example logging.DEBUG.
    Returns:
        (Logger): A logging instance with customized formatter and handlers.
    """
    level = level or logging.INFO

    logger = logging.getLogger(logger_name)
    logger.handlers = []  # reset logging handlers if they already exist

    formatter = logging.Formatter(
        fmt="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )

    if log_dir:
        logpath = os.path.join(log_dir, logger_name)
        if os.path.exists(logpath + ".log"):
            logpath += "_" + datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        logpath += ".log"

        handler = logging.FileHandler(logpath, mode="w")
        handler.setFormatter(formatter)
        handler.setLevel(level)
        logger.addHandler(handler)

    screen_handler = logging.StreamHandler(stream=sys.stdout)
    screen_handler.setFormatter(formatter)
    screen_handler.setLevel(level)
    logger.addHandler(screen_handler)

    logger.setLevel(level)
    return logger


def immutify_dictionary(d):
    """Create a frozenset-esque deterministic, unique representation of
    a nested dict.


    Args:
        d (dict): The dictionary to be immutified. Key are always strings.
            Values can be arrays of various numpy or pandas types, strings,
            numpy primitives, python native numbers, or dictionaries with
            the same format.

    Returns:
        (dict): A sorted, deterministic, unique representation of the
            dictionary.
    """
    d_new = {}
    for k, v in d.items():
        if isinstance(v, (np.ndarray, pd.Series)):
            d_new[k] = tuple(v.tolist())
        elif isinstance(v, list):
            d_new[k] = tuple(v)
        elif isinstance(v, dict):
            d_new[k] = immutify_dictionary(v)
        else:
            # convert numpy types to native
            if hasattr(v, "dtype"):
                d_new[k] = v.item()
            else:
                d_new[k] = v
    # dictionaries are ordered in python 3.6+
    return dict(sorted(d_new.items(), key=lambda item: item[0]))


def hash_dictionary(d):
    """Hash a dictionary that can be immutified with immutify_dictionary.

    Order of the keys does not matter, as dictionary becomes deterministically
    immutified. Dictionary can be nested.

    Args:
        d (dict): The dictionary to hash.

    Returns:
        (str): base16 encoded hash of the dictionary.
    """
    d_hashable = immutify_dictionary(d)
    s_hashable = json.dumps(d_hashable).encode("utf-8")
    m = hashlib.sha256(s_hashable).hexdigest()
    return m



"""
Source of ground truth for all constants used across matbench.
"""
import os

# single place for defining version
VERSION = "0.6"

# paths for metadata and validation splits
THIS_DIR = os.path.dirname(os.path.abspath(__file__))
MBV01_DATASET_METADATA_PATH = os.path.join(
    THIS_DIR, "matbench_v0.1_dataset_metadata.json"
)
MBV01_VALIDATION_DATA_PATH = os.path.join(THIS_DIR, "matbench_v0.1_validation.json")

MBV01_KEY = "matbench_v0.1"

# keys for validation files
VALIDATION_SPLIT_KEY = "splits"
VALIDATION_METADATA_KEY = "metadata"
TRAIN_KEY = "train"
TEST_KEY = "test"

# universal keys
MBID_KEY = "mbid"
REG_KEY = "regression"
CLF_KEY = "classification"
STRUCTURE_KEY = "structure"
COMPOSITION_KEY = "composition"

# scoring per task on a single fold
REG_METRICS = ["mae", "rmse", "mape", "max_error"]
CLF_METRICS = ["accuracy", "balanced_accuracy", "f1", "rocauc"]

# threshold for converting probabilities to labels
CLF_THRESH = 0.5

# scoring on a single task among folds
FOLD_DIST_METRICS = ["mean", "max", "min", "std"]



"""
Utility functions for data operations such as loading dataframes
and scoring.
"""

import logging
import math

import numpy as np
from matminer.datasets import load_dataset
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    max_error,
    mean_absolute_error,
    mean_squared_error,
    roc_auc_score,
)

from matbench.constants import (
    CLF_KEY,
    CLF_METRICS,
    CLF_THRESH,
    MBID_KEY,
    REG_KEY,
    REG_METRICS,
)
from matbench.metadata import mbv01_metadata

logger = logging.getLogger(__name__)


def load(dataset_name, dataset_metadata=mbv01_metadata):
    """
    Load a matbench dataset into memory as a dataframe.
    This function is simply a wrapper around the matminer dataloader.

    Each matbench dataset is completely self contained.

    See https://hackingmaterials.lbl.gov/matbench/ for a list of dataset names.
    For example, "matbench_jdft2d".

    Args:
        dataset_name (str): A matbench dataset name, as defined in the
        datasets.json.

    Returns:
        (pandas.DataFrame): The dataset, containing two columns:
            - Inputs, either compositions or pymatgen structure objects.
            - Outputs, either a float (for regression) or a boolean (for
                classification).
    """

    if dataset_name not in dataset_metadata:
        raise KeyError(
            f"Dataset name {dataset_name} not recognized by matbench. "
            f"Please see https://hackingmaterials.lbl.gov/matbench for "
            f"a list of the dataset names, or choose from:"
            f"\n{list(dataset_metadata.keys())}"
        )
    logger.debug(
        f"Loading {dataset_name} into memory; please be patient as "
        f"loading many structures can take a while to serialize."
    )
    df = load_dataset(dataset_name)

    id_n_zeros = math.floor(math.log(df.shape[0], 10)) + 1
    mpcontribs_prefix = dataset_name.replace("matbench", "mb").replace("_", "-")
    df[MBID_KEY] = [f"{mpcontribs_prefix}-{i + 1:0{id_n_zeros}d}" for i in df.index]

    df = df.set_index(MBID_KEY)
    df = df[
        [
            dataset_metadata[dataset_name].input_type,
            dataset_metadata[dataset_name].target,
        ]
    ]

    return df


def score_array(true_array, pred_array, task_type):
    """
    Score an array according to multiple metrics.

    Args:
        true_array (list or np.array): The ground truth array
        pred_array (list or np.array): The predicted (test) array
        task_type (str): Either regression or classification.

    Returns:
        (dict): dictionary of the scores, according to all defined
            metrics.

    """
    computed = {}

    if task_type == REG_KEY:
        metrics = REG_METRICS
    elif task_type == CLF_KEY:
        metrics = CLF_METRICS
    else:
        raise ValueError(
            f"'task_type' must be one of {[REG_KEY, CLF_KEY]}, not '{task_type}'"
        )

    for metric in metrics:
        mfunc = METRIC_MAP[metric]

        if metric == "rocauc":
            # Both arrays must be in probability form
            # if pred. array is given in probabilities
            if isinstance(pred_array[0], float):
                true_array = homogenize_clf_array(true_array, to_probs=True)

        # Other clf metrics always be converted to labels
        elif metric in CLF_METRICS:
            if isinstance(pred_array[0], float):
                pred_array = homogenize_clf_array(pred_array, to_labels=True)

        computed[metric] = mfunc(true_array, pred_array)
    return computed


def mean_absolute_percentage_error(y_true, y_pred, threshold=1e-5):
    """Compute mean absolute percentage error, masked

    Masking is for when y_true is zero (causing a
    divide by zero error) or when y_true is very small
    (causing a massive skewing in the absolute percentage
    error).

    **Note: THIS WILL IGNORE ALL ENTRIES WHERE y_true's
    MAGNITUDE IS less than the threshold, hence the
    MAPE score is not representative of all
    entries if the truth array contains entries with
    magnitude very close to 0.**

    Args:
        y_true (np.ndarray): A 1-D array of true values
        y_pred (np.ndarray): A 1-D array of predicted values
        threshold (float): Entries with magnitude below this
            value will be ignored in the output.

    Returns:

    """
    y_true = np.asarray(y_true)
    mask = np.abs(y_true) > threshold
    y_pred = np.asarray(y_pred)
    y_true = y_true[mask]
    y_pred = y_pred[mask]
    return np.mean(np.fabs((y_true - y_pred) / y_true))


def homogenize_clf_array(array, to_probs=False, to_labels=False, thresh=CLF_THRESH):
    """
    Homogenize an array of either:

    1. labels (True, False) to probabilities (1.0, 0.0)
    2. probabilities (between 0 and 1) to labels (True, False)
        based on a threshold float

    Args:
        array ([bool], [float]): A list of bools or a list of floats 0-1.
        to_probs (bool): Convert the input array to all probabilities
        to_labels (bool): Convert the input array to all labels based on
            the threshold value thresh.
        thresh (float): A number 0-1, which will decide the threshold
            of probabilities if to_labels is True

    Returns:
        list
    """
    if sum([to_probs, to_labels]) != 1:
        raise ValueError(
            "Set ONE of to_probs or to_labels to True to define "
            "the conversion, NOT both."
        )

    if to_probs:
        if all([isinstance(i, bool) for i in array]):
            # The source array is bools
            homogenized = [1.0 if i is True else 0.0 for i in array]
            return homogenized
        else:
            raise TypeError(
                "Cannot convert non-bool type in clf array to " "probabilities."
            )
    elif to_labels:
        if all([isinstance(i, float) for i in array]):
            # The source array is probabilities
            homogenized = np.asarray(array) > thresh
            return homogenized.tolist()
        else:
            raise TypeError(
                "Cannot convert non-float types in clf array to" "labels."
            )


METRIC_MAP = {
    "mae": mean_absolute_error,
    "rmse": lambda true, pred: math.sqrt(mean_squared_error(true, pred)),
    "mape": mean_absolute_percentage_error,
    "max_error": max_error,
    "accuracy": accuracy_score,
    "balanced_accuracy": balanced_accuracy_score,
    "f1": f1_score,
    "rocauc": roc_auc_score,
}



from plotly.io import read_json






# fig = read_json("/Users/ardunn/alex/lbl/projects/common_env/dev_codes/matbench/docs_src/static/scaled_errors.json")
fig = read_json("/Users/ardunn/alex/lbl/projects/common_env/dev_codes/matbench/docs_src/static/task_matbench_v0.1_matbench_steels.json")

fig.write_html("./gap_expt.html")



import copy
import glob
import json
import os
import unittest
from warnings import warn

import numpy as np
import pandas as pd
from pymatgen.core import Structure

from matbench.constants import (
    CLF_METRICS,
    COMPOSITION_KEY,
    FOLD_DIST_METRICS,
    MBV01_KEY,
    REG_KEY,
    REG_METRICS,
    STRUCTURE_KEY,
    TEST_KEY,
)
from matbench.task import MatbenchTask
from matbench.tests.util import TEST_DIR, model_random


class TestMatbenchTask(unittest.TestCase):
    test_datasets = ("matbench_dielectric", "matbench_steels", "matbench_glass")

    def tearDown(self) -> None:
        # remove all temporary output files
        for f in glob.glob(os.path.join(TEST_DIR, "*_output.json")):
            os.remove(f)

    def test_instantiation(self):
        for ds in self.test_datasets:
            MatbenchTask(ds, autoload=True)

    def test_get_train_and_val_data(self):
        # Assuming 5-fold nested cross validation
        for ds in self.test_datasets:
            mbt = MatbenchTask(ds, autoload=False)
            mbt.load()
            # shuffle seed must be set because it shuffles
            # the (same) training data in a deterministic manner

            inputs, outputs = mbt.get_train_and_val_data(
                fold_number=0, as_type="tuple"
            )

            self.assertListEqual(inputs.index.tolist(), outputs.index.tolist())
            self.assertEqual(inputs.shape[0], int(np.floor(mbt.df.shape[0] * 4 / 5)))
            self.assertEqual(
                outputs.shape[0], int(np.floor(mbt.df.shape[0] * 4 / 5))
            )

            input_type = (
                Structure if mbt.metadata.input_type == STRUCTURE_KEY else str
            )
            output_type = float if mbt.metadata.task_type == REG_KEY else bool
            self.assertTrue(all([isinstance(d, input_type) for d in inputs]))
            self.assertTrue(all([isinstance(d, output_type) for d in outputs]))

            if ds == "matbench_dielectric":
                mat = inputs.loc["mb-dielectric-1985"]
                f = mat.composition.reduced_formula
                val = outputs.loc["mb-dielectric-1985"]
                self.assertEqual(f, "Re3(TeBr)7")
                self.assertEqual(
                    inputs.iloc[1621], mat
                )  # ensure the ordering is correct via iloc
                n = 2.5230272821931656
                self.assertAlmostEqual(val, n, places=10)
                self.assertAlmostEqual(outputs.iloc[1621], n, places=10)
            elif ds == "matbench_steels":
                alloy = (
                    "Fe0.692C0.00968Mn0.000101Si0.0144Cr0.133Ni0.00887"
                    "Mo0.0114V0.000109Nb0.000477Co0.130Al0.000616"
                )
                mat = inputs.loc["mb-steels-095"]
                val = outputs.loc["mb-steels-095"]
                self.assertEqual(alloy, mat)
                self.assertEqual(alloy, inputs.iloc[75])
                yield_strength = 1369.5
                self.assertAlmostEqual(val, yield_strength, places=5)
                self.assertAlmostEqual(outputs.iloc[75], yield_strength, places=5)
            elif ds == "matbench_glass":
                alloy = "Ce2Al5Cu43"
                mat = inputs.loc["mb-glass-0600"]
                val = outputs.loc["mb-glass-0600"]
                self.assertEqual(alloy, mat)
                self.assertEqual(alloy, inputs.iloc[480])
                gfa = False
                self.assertEqual(val, gfa)
                self.assertEqual(outputs.iloc[480], gfa)

    def test_get_test_data(self):
        for ds in self.test_datasets:
            mbt = MatbenchTask(ds, autoload=False)
            mbt.load()
            folds = []
            for fold in mbt.folds:
                inputs, outputs = mbt.get_test_data(
                    fold_number=fold, as_type="tuple", include_target=True
                )

                self.assertListEqual(inputs.index.tolist(), outputs.index.tolist())

                upper_bound = int(np.ceil(mbt.df.shape[0] / 5))
                allowed_fold_sizes = (upper_bound - 1, upper_bound)
                self.assertTrue(inputs.shape[0] in allowed_fold_sizes)
                self.assertTrue(outputs.shape[0] in allowed_fold_sizes)
                input_type = (
                    Structure if mbt.metadata.input_type == STRUCTURE_KEY else str
                )
                output_type = float if mbt.metadata.task_type == REG_KEY else bool
                self.assertTrue(all([isinstance(d, input_type) for d in inputs]))
                self.assertTrue(all([isinstance(d, output_type) for d in outputs]))
                folds.append((inputs, outputs))

                df_test = mbt.get_test_data(
                    fold_number=fold, as_type="df", include_target=True
                )
                self.assertListEqual(
                    df_test.columns.tolist(),
                    [mbt.metadata.input_type, mbt.metadata.target],
                )

                df_test = mbt.get_test_data(
                    fold_number=fold, as_type="df", include_target=False
                )
                self.assertTrue(isinstance(df_test, pd.DataFrame))
                self.assertListEqual(
                    df_test.columns.tolist(), [mbt.metadata.input_type]
                )

            # check if all entries from original df are in exactly
            # one test fold exactly once
            original_input_df = mbt.df[mbt.metadata.input_type]
            inputs_from_folds = pd.concat([f[0] for f in folds])
            self.assertEqual(inputs_from_folds.shape[0], original_input_df.shape[0])
            self.assertTrue(
                original_input_df.apply(
                    lambda i: i in inputs_from_folds.tolist()
                ).all()
            )

            # Test individual samples from an individual test set
            inputs, outputs = folds[0]
            if ds == "matbench_dielectric":
                ki = inputs.iloc[12].composition.reduced_formula
                self.assertEqual(ki, "KI")
                self.assertEqual(
                    ki, inputs.loc["mb-dielectric-0076"].composition.reduced_formula
                )
                self.assertEqual(
                    ki,
                    mbt.df[STRUCTURE_KEY]
                    .loc["mb-dielectric-0076"]
                    .composition.reduced_formula,
                )
                n = 1.7655027612552967
                self.assertAlmostEqual(outputs.iloc[12], n, places=10)
                self.assertAlmostEqual(
                    outputs.loc["mb-dielectric-0076"], n, places=10
                )
                self.assertAlmostEqual(
                    mbt.df[mbt.metadata.target].loc["mb-dielectric-0076"],
                    n,
                    places=10,
                )
            elif ds == "matbench_steels":
                alloy = (
                    "Fe0.682C0.00877Mn0.000202Si0.00967Cr0.134"
                    "Ni0.00907Mo0.00861V0.00501Nb0.0000597Co0.142Al0.000616"
                )
                self.assertEqual(alloy, inputs.loc["mb-steels-068"])
                self.assertEqual(alloy, inputs.iloc[12])
                self.assertEqual(alloy, mbt.df[COMPOSITION_KEY].loc["mb-steels-068"])
                yield_strength = 1241.0
                self.assertAlmostEqual(outputs.iloc[12], yield_strength, places=5)
                self.assertAlmostEqual(
                    outputs.loc["mb-steels-068"], yield_strength, places=5
                )
                self.assertAlmostEqual(
                    mbt.df[mbt.metadata.target].loc["mb-steels-068"],
                    yield_strength,
                    places=5,
                )
            elif ds == "matbench_glass":
                alloy = "Al13VCu6"
                self.assertEqual(alloy, inputs.iloc[12])
                self.assertEqual(alloy, inputs.loc["mb-glass-0056"])
                self.assertEqual(alloy, mbt.df[COMPOSITION_KEY].loc["mb-glass-0056"])
                gfa = True
                self.assertEqual(outputs.iloc[12], gfa)
                self.assertEqual(outputs.loc["mb-glass-0056"], gfa)
                self.assertEqual(
                    mbt.df[mbt.metadata.target].loc["mb-glass-0056"], gfa
                )

    def test_get_info(self):
        mbt = MatbenchTask("matbench_steels", autoload=False)
        mbt.get_info()
        self.assertTrue("citations" in mbt.info.lower())
        self.assertTrue("SHA256 Hash Digest" in mbt.info)

    def test_record(self):
        self._test_record()

    def test_record_std(self):
        # ensure that conversion from std to ci is correct
        mbt = MatbenchTask(self.test_datasets[0], autoload=True)
        _, test_outputs = mbt.get_test_data(0, as_type="tuple", include_target=True)
        n_test = len(test_outputs)
        mbt.record(0, [0.0] * n_test, std=[1.0] * n_test)
        ci_0_0 = list(mbt.results["fold_0"]["uncertainty"].items())[0][1]
        ci_lower = ci_0_0["ci_lower"]
        ci_upper = ci_0_0["ci_upper"]
        # mean:0 std:1, 95% CI-->c.a. 1.96
        # https://stackoverflow.com/a/29562808/13697228
        ci_upp_check = 1.959963984540054
        ci_low_check = -ci_upp_check
        msg = f"""Conversion from normal distribution standard deviation to
        lower 95% confidence bound ({ci_upper}) is not within tolerance of
        expected confidence bound ({ci_upp_check})"""
        self.assertAlmostEqual(ci_upper, ci_upp_check, msg=msg)
        msg = f"""Conversion from normal distribution standard deviation
        to upper 95% confidence bound ({ci_lower}) is not within tolerance
        of expected confidence bound ({ci_low_check})"""
        self.assertAlmostEqual(ci_lower, ci_low_check, msg=msg)

        # carry out test as normal
        self._test_record(uq_type="std")

    def test_record_ci(self):
        # ensure that conversion from ci to std is correct
        mbt = MatbenchTask(self.test_datasets[0], autoload=True)
        _, test_outputs = mbt.get_test_data(0, as_type="tuple", include_target=True)
        n_test = len(test_outputs)
        ci_upper = 1.959963984540054
        ci_lower = -ci_upper
        mbt.record(0, [0.0] * n_test, ci=[[ci_lower, ci_upper]] * n_test)
        ci_0_0 = list(mbt.results["fold_0"]["uncertainty"].items())[0][1]
        std = ci_0_0["std"]
        std_check = 1.0
        # mean:0 std:1, 95% CI-->c.a. 1.96
        # https://stackoverflow.com/a/29562808/13697228
        msg = f"""Conversion from symmetric 95% confidence bounds to
        standard deviation ({std}) is not within tolerance of expected
        confidence bound ({std_check})"""
        self.assertAlmostEqual(std, std_check, msg=msg)

        self._test_record(uq_type="ci")

    def _test_record(self, uq_type=None):
        for ds in self.test_datasets:
            if ds == "matbench_glass":
                uq_type = None
                warn(
                    """Overriding uq_type as None due to incompatibility
                     with classification tasks (expected behavior)."""
                )
            # Testing two scenarios: model is perfect, and model is random
            for model_is_perfect in (True, False):
                mbt = MatbenchTask(ds, autoload=False)
                mbt.load()

                # test to make sure raw data output is correct,
                # using a random model
                for fold, fold_key in mbt.folds_map.items():
                    _, training_outputs = mbt.get_train_and_val_data(
                        fold, as_type="tuple"
                    )
                    if model_is_perfect:
                        test_inputs, test_outputs = mbt.get_test_data(
                            fold, as_type="tuple", include_target=True
                        )
                        model_response = test_outputs
                    else:
                        test_inputs = mbt.get_test_data(
                            fold, as_type="tuple", include_target=False
                        )
                        model_response = model_random(
                            training_outputs,
                            test_inputs,
                            response_type=mbt.metadata.task_type,
                        )

                    n_test = len(test_inputs)

                    # uncertainty quantification parameter
                    uq_param = {}
                    if uq_type == "ci":
                        uq_shape = (n_test, 2)
                    elif uq_type == "std":
                        uq_shape = (n_test,)
                    if uq_type is not None:
                        uq_param[uq_type] = 2 * np.random.rand(*uq_shape)

                    dummy_params = {
                        "test_param": 1,
                        "other_param": "string",
                        "hyperparam": True,
                    }

                    mbt.record(
                        fold,
                        predictions=model_response,
                        params=dummy_params,
                        **uq_param,
                    )
                    self.assertEqual(
                        len(mbt.results[fold_key].data.values()), n_test
                    )
                    self.assertEqual(mbt.results[fold_key].parameters.test_param, 1)
                    self.assertEqual(
                        mbt.results[fold_key].parameters.other_param, "string"
                    )
                    self.assertEqual(
                        mbt.results[fold_key].parameters.hyperparam, True
                    )

                if ds == "matbench_dielectric":
                    mae = mbt.results.fold_0.scores.mae
                    val = mbt.results.fold_0.data["mb-dielectric-0008"]
                    if model_is_perfect:
                        self.assertAlmostEqual(mae, 0.0, places=10)
                        self.assertAlmostEqual(val, 2.0323401126123875, places=10)
                    else:
                        self.assertAlmostEqual(mae, 29.790913986352297, places=10)
                        self.assertAlmostEqual(val, 43.36354273040313, places=10)
                elif ds == "matbench_steels":
                    mae = mbt.results.fold_0.scores.mae
                    if model_is_perfect:
                        self.assertAlmostEqual(mae, 0.0, places=10)
                    else:
                        self.assertAlmostEqual(mae, 488.97286237333986, places=10)
                elif ds == "matbench_glass":
                    rocauc = mbt.results.fold_0.scores.rocauc
                    if model_is_perfect:
                        self.assertAlmostEqual(rocauc, 1.0, places=10)
                    else:
                        self.assertAlmostEqual(rocauc, 0.5141975796883651, places=10)

                self.assertTrue(mbt.all_folds_recorded)

            mbt = MatbenchTask(self.test_datasets[0], autoload=True)
            # Test to make sure bad predictions won't be recorded
            with self.assertRaises(ValueError):
                mbt.record(0, [0.0, 1.0])

            with self.assertRaises(ValueError):
                mbt.record(0, ["not", "a number"])

    def test_MSONability(self):
        for ds in self.test_datasets:
            mbt = MatbenchTask(ds, autoload=False)
            mbt.load()

            for fold in mbt.folds:
                _, training_outputs = mbt.get_train_and_val_data(
                    fold, as_type="tuple"
                )
                test_inputs, test_outputs = mbt.get_test_data(
                    fold, as_type="tuple", include_target=True
                )
                mbt.record(
                    fold,
                    predictions=test_outputs,
                    params={"some_param": 1, "another param": 30349.4584},
                )

            d = mbt.as_dict()

            self.assertEqual(d["@module"], "matbench.task")
            self.assertEqual(d["@class"], "MatbenchTask")
            self.assertEqual(d[mbt._BENCHMARK_KEY], MBV01_KEY)
            self.assertEqual(d[mbt._DATASET_KEY], ds)
            self.assertEqual(len(d["results"]), len(mbt.validation.keys()))

            for fold, fold_key in mbt.folds_map.items():
                res = d["results"][fold_key]
                self.assertIn(MatbenchTask._PARAMS_KEY, res)
                self.assertIn(MatbenchTask._SCORES_KEY, res)
                self.assertIn(MatbenchTask._DATA_KEY, res)

                # make sure test set as per MbT and the recorded predictions
                # are the same shape inside dict
                self.assertEqual(
                    len(res["data"]), len(mbt.validation[fold_key][TEST_KEY])
                )

            mbt.to_file(os.path.join(TEST_DIR, f"msonability_{ds}_output.json"))

            # todo: uncomment to regenerate test files
            # todo: these can be used as the score_matbench_*_perfect.json
            # todo: files as well if renamed.
            # mbt.to_file(os.path.join(TEST_DIR, f"msonability_{ds}.json"))
            # mbt.to_file(os.path.join(TEST_DIR, f"scores_{ds}_perfect.json"))

            # Test ingestion from ground truth json files
            truth_fname = os.path.join(TEST_DIR, f"msonability_{ds}.json")

            with open(truth_fname, "r") as f:
                truth = json.loads(f.read())
            MatbenchTask.from_file(truth_fname)
            MatbenchTask.from_dict(truth)

            # Ensure errors are thrown for bad json

            missing_results = copy.deepcopy(truth)
            missing_results.pop("results")

            with self.assertRaises(KeyError):
                MatbenchTask.from_dict(missing_results)

            for key in [
                MatbenchTask._PARAMS_KEY,
                MatbenchTask._DATA_KEY,
                MatbenchTask._SCORES_KEY,
            ]:
                missing_key = copy.deepcopy(truth)
                missing_key["results"]["fold_3"].pop(key)

                with self.assertRaises((ValueError, KeyError)):
                    MatbenchTask.from_dict(missing_key)

            # If an otherwise perfect json is missing a required index
            missing_ix_fold0 = copy.deepcopy(truth)

            missing_ix = mbt.validation.fold_0.test[0]
            missing_ix_fold0["results"]["fold_0"]["data"].pop(missing_ix)

            with self.assertRaises(ValueError):
                MatbenchTask.from_dict(missing_ix_fold0)

            # If an otherwise perfect json has an extra index
            extra_ix_fold0 = copy.deepcopy(truth)

            extra_viable_ix = mbt.validation.fold_0.train[0]
            extra_ix_fold0["results"]["fold_0"]["data"][extra_viable_ix] = 1.92

            with self.assertRaises(ValueError):
                MatbenchTask.from_dict(extra_ix_fold0)

            # If an otherwise perfect json has a wrong datatype
            wrong_dtype = copy.deepcopy(truth)
            wrong_dtype["results"]["fold_2"]["data"][
                mbt.validation.fold_2.test[4]
            ] = "any string"

            with self.assertRaises(TypeError):
                MatbenchTask.from_dict(wrong_dtype)

    def test_autoload(self):
        mbt = MatbenchTask("matbench_steels", autoload=False)
        with self.assertRaises(ValueError):
            mbt._check_is_loaded()

        with self.assertRaises(ValueError):
            mbt.get_test_data(0)

        with self.assertRaises(ValueError):
            mbt.get_train_and_val_data(0)

        mbt.load()
        mbt._check_is_loaded()
        mbt.get_test_data(0)
        mbt.get_train_and_val_data(0)

        MatbenchTask("matbench_steels", autoload=True)

    def test_scores(self):
        mbt = MatbenchTask.from_file(
            os.path.join(TEST_DIR, "scores_matbench_dielectric_perfect.json")
        )

        for metric in REG_METRICS:
            for fdm in FOLD_DIST_METRICS:
                self.assertAlmostEqual(0.0, mbt.scores[metric][fdm], places=10)

        mbt = MatbenchTask.from_file(
            os.path.join(TEST_DIR, "scores_matbench_glass_perfect.json")
        )

        for metric in CLF_METRICS:
            for fdm in FOLD_DIST_METRICS:
                test_val = 0.0 if fdm == "std" else 1.0
                self.assertAlmostEqual(test_val, mbt.scores[metric][fdm], places=10)

    def test_usage(self):
        # access some common attrs
        mbt_clf = MatbenchTask.from_file(
            os.path.join(TEST_DIR, "scores_matbench_dielectric_perfect.json")
        )
        mbt_reg = MatbenchTask.from_file(
            os.path.join(TEST_DIR, "scores_matbench_glass_perfect.json")
        )

        for mbt in (mbt_clf, mbt_reg):
            for index, datum in mbt.results.fold_2.data.items():
                self.assertTrue(isinstance(datum, (bool, float)))
                self.assertTrue("mb-" in index)

        self.assertTrue(
            isinstance(mbt.results.fold_3.parameters, (dict, type(None)))
        )

    def test_clf_conversion(self):
        mbt = MatbenchTask("matbench_glass", autoload=True)

        for fold in mbt.folds:
            trdf = mbt.get_test_data(fold, as_type="df")
            pred = np.random.random(len(trdf))
            mbt.record(fold, pred)

        # With random sampling it should almost never exceed
        # ROCAUC of 0.6
        self.assertLess(mbt.scores["rocauc"]["mean"], 0.6)

        # Try the same model as perfect
        mbt = MatbenchTask("matbench_glass", autoload=True)

        for fold in mbt.folds:
            inputs, targets = mbt.get_test_data(fold, include_target=True)
            pred = [1.0 if i else 0.0 for i in targets]
            mbt.record(fold, pred)

        self.assertAlmostEqual(mbt.scores["rocauc"]["mean"], 1.0, places=5)
        self.assertAlmostEqual(mbt.scores["rocauc"]["std"], 0.0, places=5)

    def test_has_polymorphs(self):
        # Test a composition input type
        mbt = MatbenchTask("matbench_steels", autoload=True)
        self.assertFalse(mbt.has_polymorphs)

        # Test similarly for a composition dataframe with polymorphs
        df_comp_poly = pd.DataFrame(
            {"composition": ["Al2O3", "Cr2O3", "Al4O6"], "property": [1.0, 1.1, 1.2]}
        )
        mbt.df = df_comp_poly
        self.assertTrue(mbt.has_polymorphs)

        # Test a structure input type
        mbt = MatbenchTask("matbench_dielectric", autoload=True)
        self.assertTrue(mbt.has_polymorphs)

        # Test similarly for a structure dataframe without polymorphs
        mbt.df = mbt.df.iloc[:10]
        self.assertFalse(mbt.has_polymorphs)



import copy
import logging

import numpy as np
from matminer.datasets import get_all_dataset_info
from matminer.featurizers.conversions import (
    StrToComposition,
    StructureToComposition,
)
from monty.json import MSONable
from scipy import stats

from matbench.constants import (
    CLF_KEY,
    CLF_METRICS,
    FOLD_DIST_METRICS,
    MBV01_KEY,
    REG_KEY,
    REG_METRICS,
)
from matbench.data_ops import load, score_array
from matbench.metadata import mbv01_metadata, mbv01_validation
from matbench.util import MSONable2File, RecursiveDotDict, immutify_dictionary

logger = logging.getLogger(__name__)


class MatbenchTask(MSONable, MSONable2File):
    """The core interface for running a Matbench task and recording its results.

    MatbenchTask handles creating training/validation and testing sets, as
    well as recording and managing all data in a consistent fashion.
    MatbenchTask also validates data according to te specifications in the
    validation file.

    MatbenchTasks have a few core methods:

    - MatbenchTask.get_train_and_val_data: Get nested cross validation data to
        be used for all training and validation.
    - MatbenchTask.get_test_data: Get test data for nested cross validation.
    - MatbenchTask.record: Record your predicted results for the test data.
    - MatbenchTask.validate: Check to make sure the data you recorded for this
        task is valid.

    You can iterate through the folds of a matbench task using .folds and
    the .get_*_data methods.

    You can load the results of a task without having to load large
    datasets themselves. However, to get training and testing data,
    you must load the datasets. Tasks loaded from files do not
    automatically load the dataset into memory; to load a dataset into memory,
    use MatbenchTask.load().

    See the full documentation online for more info and tutorials on
    using MatbenchTask.

    Attributes:
        benchmark_name (str): The name of the benchmark this task belongs to.
        df (pd.DataFrame): the dataframe of the dataset for this task
        info (str): Info about this dataset
        metadata (RecursiveDotDict): all metadata about this dataset
        validation (RecursiveDotDict): The validation specification for this
            task, including the training and testing splits for each fold.
        folds_keys ([str]): Keys of folds, fold_i for the ith fold.
        folds_nums ([int]): Values of folds, i for the ith fold.
        folds_map ({int: str}): Mapping of folds_nums to folds_keys
        folds ([int]): Alias for folds_nums
        results (RecursiveDotDict): all raw results in dict-like form.
    """

    _RESULTS_KEY = "results"
    _BENCHMARK_KEY = "benchmark_name"
    _DATASET_KEY = "dataset_name"
    _DATA_KEY = "data"
    _UNCERTAINTY_KEY = "uncertainty"
    _PARAMS_KEY = "parameters"
    _SCORES_KEY = "scores"

    def __init__(self, dataset_name, autoload=True, benchmark=MBV01_KEY):
        """
        Args:
            dataset_name (str): Name of the task. Must belong to the benchmark
                given in the 'benchmark' argument.
            autoload (bool): If True, will load the benchmark's raw data. This
                includes deserializing many large structures for some datasets,
                so loading make take some time. If False, you will need to
                run .load() before running .get_*_data() methods.
            benchmark (str): Name of the benchmark this task belongs to.
        """
        self.dataset_name = dataset_name
        self.df = load(self.dataset_name) if autoload else None
        self.info = get_all_dataset_info(dataset_name)

        # define all static data needed for this task
        # including citations, data size, as well as specific validation splits

        if benchmark == MBV01_KEY:
            self.benchmark_name = MBV01_KEY
            self.metadata = mbv01_metadata[dataset_name]
            self.validation = mbv01_validation.splits[dataset_name]
        else:
            raise ValueError(
                f"Only {MBV01_KEY} available. No other benchmarks defined!"
            )

        # keeping track of folds
        self.folds_keys = list(self.validation.keys())
        self.folds_nums = list(range(len(self.folds_keys)))
        self.folds_map = dict(zip(self.folds_nums, self.folds_keys))

        # Alias for ease of use
        self.folds = self.folds_nums
        self.results = RecursiveDotDict({})

    def __repr__(self) -> str:
        keys = "input_type mad n_samples target task_type unit".split()

        md_str = ",\n  ".join(f"{k}={self.metadata[k]}" for k in keys)

        return (
            f"{type(self).__name__}(\n  dataset_name={self.dataset_name},\n  version"
            f"={self.benchmark_name.replace('matbench_v', '')},\n  {md_str},\n)"
        )

    def _get_data_from_df(self, ids, as_type):
        """Private function to get fold data from the task dataframe.

        Args:
            ids (list-like): List of string indices to grab from the df.
            as_type (str): either "df" or "tuple". If "df", returns the
                data as a subset of the task df. If "tuple", returns
                list-likes of the inputs and outputs as a 2-tuple.

        Returns:
            (pd.DataFrame or (list-like, list-like))

        """
        relevant_df = self.df.loc[ids]
        if as_type == "df":
            return relevant_df
        elif as_type == "tuple":
            # inputs, outputs
            return (
                relevant_df[self.metadata.input_type],
                relevant_df[self.metadata.target],
            )

    def _check_is_loaded(self):
        """Private method to check if the dataset is loaded.

        Throws error if the dataset is not loaded.

        Returns:
            None
        """
        if self.df is None:
            raise ValueError(
                "Task dataset is not loaded! Run MatbenchTask.load() to "
                "load the dataset into memory."
            )

    def _check_all_folds_recorded(self, msg):
        """Private method to check if all folds have been recorded.

        Throws error if all folds have not been recorded.

        Args:
            msg (str): Error message to be displayed.

        Returns:
            None
        """
        if not self.all_folds_recorded:
            raise ValueError(
                f"{msg}; folds "
                f"{[f for f in self.is_recorded if not self.is_recorded[f]]} "
                f"not recorded!"
            )

    @classmethod
    def from_dict(cls, d):
        """Create a MatbenchTask from a dictionary input.

        Required method from MSONable.

        Args:
            d (dict):

        Returns:
            (MatbenchTask)

        """
        req_base_keys = [
            "@module",
            "@class",
            cls._DATASET_KEY,
            cls._RESULTS_KEY,
            cls._BENCHMARK_KEY,
        ]
        for k in req_base_keys:
            if k not in d:
                raise KeyError(f"Required key '{k}' not found.")
        extra_base_keys = [k for k in d.keys() if k not in req_base_keys]
        if extra_base_keys:
            raise KeyError(f"Extra keys {extra_base_keys} not allowed.")
        return cls._from_args(
            dataset_name=d[cls._DATASET_KEY],
            benchmark_name=d[cls._BENCHMARK_KEY],
            results_dict=d[cls._RESULTS_KEY],
        )

    @classmethod
    def _from_args(cls, dataset_name, benchmark_name, results_dict):
        """Instantiate a MatbenchTask from a arguments

        Args:
            dataset_name (str): The name of the dataset/task
            benchmark_name (str): The name of the corresponding benchmark
            results_dict (dict): A formatted dictionary of raw results.

        Returns:
            (MatbenchTask)
        """
        obj = cls(dataset_name, autoload=False, benchmark=benchmark_name)
        obj.results = RecursiveDotDict(results_dict)
        obj.validate()
        return obj

    def load(self):
        """Load the dataset for this task into memory.

        Returns:
            None
        """
        if self.df is None:
            logger.info(f"Loading dataset '{self.dataset_name}'...")
            self.df = load(self.dataset_name)
            logger.info(f"Dataset '{self.dataset_name} loaded.")
        else:
            logger.info(
                f"Dataset {self.dataset_name} already loaded; "
                f"not reloading dataset."
            )

    def get_info(self):
        logger.info(self.info)

    def get_train_and_val_data(self, fold_number, as_type="tuple"):
        """
        The training + validation data. All model tuning and
        hyperparameter selection must be done on this data, NOT test data.

        Args:
            fold_number:

        Returns:

        """
        self._check_is_loaded()
        fold_key = self.folds_map[fold_number]
        ids = self.validation[fold_key].train
        return self._get_data_from_df(ids, as_type)

    def get_test_data(self, fold_number, as_type="tuple", include_target=False):
        """
        The test data used for recording benchmarks.

        Args:
            fold_number:

        Returns:


        """
        self._check_is_loaded()
        fold_key = self.folds_map[fold_number]
        ids = self.validation[fold_key].test
        if include_target:
            return self._get_data_from_df(ids, as_type)
        else:
            if as_type == "tuple":
                return self._get_data_from_df(ids, as_type)[0]
            elif as_type == "df":
                return self._get_data_from_df(ids, as_type)[
                    [self.metadata.input_type]
                ]

    def record(self, fold_number, predictions, ci=None, std=None, params=None):
        """Record the test data as well as parameters about the model
        trained on this fold.

        Args:
            fold_number (int): The fold number.
            predictions ([float] or [bool] or np.ndarray): A list of predictions for
            fold number {fold_number}
            ci ([tuple] or [list] or np.ndarray): A list of 95% confidence
            intervals on predictions for fold number {fold_number}. By default
            None. Only one of `ci` or `std` should be specified, not both.
            std ([float] or np.ndarray): A list of prediction standard deviations
            for fold number {fold_number}. By default None. Only one of
            `ci` or `std` should be specified, not both.
            params (dict): Any free-form parameters for information
                about the algorithm on this fold. For example,
                hyperparameters determined during validation. Parameters
                must be a dictionary; dictionary types must adhere to
                the same requirements as in the MatbenchBenchmark.add_metadata
                docstring.

        Returns:
            None
        """
        if self.is_recorded[fold_number]:
            logger.error(
                f"Fold number {fold_number} already recorded! Aborting record..."
            )
        else:
            # avoid problems with json serialization
            if isinstance(predictions, np.ndarray):
                predictions = predictions.tolist()

            if isinstance(std, np.ndarray):
                std = std.tolist()

            if isinstance(ci, np.ndarray):
                ci = ci.tolist()

            if std is not None and ci is not None:
                raise ValueError(
                    """Both standard deviation (`std`) and confidence
                    intervals (`ci`) were specified as kwargs. Only one
                    should be specified, not both."""
                )

            fold_key = self.folds_map[fold_number]

            # create map of original df index to prediction, e.g.,
            # {ix_of_original_df1: prediction1, ... etc.}

            split_ids = self.validation[fold_key].test
            if len(predictions) != len(split_ids):
                raise ValueError(
                    f"Prediction outputs must be the same length as the "
                    f"inputs! {len(predictions)} != {len(split_ids)}"
                )

            ids_to_predictions = {split_ids[i]: p for i, p in enumerate(predictions)}
            self.results[fold_key][self._DATA_KEY] = ids_to_predictions

            if std is not None or ci is not None:
                if self.metadata["task_type"] == "classification":
                    raise ValueError(
                        "`std` and `ci` are not valid kwargs for classification "
                        + "tasks. See "
                        + "https://github.com/materialsproject/matbench/pull/99/files#issuecomment-1022662192."  # noqa: E501
                    )

                if ci is None:
                    low_p = 0.05
                    high_p = 0.95
                    # convert from two-tail to one-tail probabilities
                    # for compatibility with `ppf`
                    # https://stackoverflow.com/a/29562808/13697228
                    low_p = low_p / 2.0
                    high_p = (1 + high_p) / 2.0
                    # convert std to ci, modified from source:
                    # https://github.com/uncertainty-toolbox/uncertainty-toolbox/blob/b2f342f6606d1d667bf9583919a663adf8643efe/uncertainty_toolbox/metrics_scoring_rule.py#L187 # noqa: E501
                    pred_l = stats.norm.ppf(low_p, loc=predictions, scale=std)
                    pred_u = stats.norm.ppf(high_p, loc=predictions, scale=std)
                    ci = np.vstack((pred_l.ravel(), pred_u.ravel())).T.tolist()
                    ci = [tuple(c) for c in ci]

                if std is None:
                    # std calculated and stored iff ci is symmetric within tol
                    pred_l, pred_u = np.hsplit(np.array(ci), 2)
                    if np.allclose(-pred_l, pred_u):
                        high_p = 0.95
                        # convert from two-tail to one-tail probabilities for
                        # compatibility with `ppf`
                        # https://stackoverflow.com/a/29562808/13697228
                        high_p = (1 + high_p) / 2.0
                        std = (pred_u - pred_l) / (2 * stats.norm.ppf(high_p))
                    else:
                        std = [None] * len(ci)

                if len(ci) != len(split_ids):
                    raise ValueError(
                        f"""Confidence interval outputs (derived from standard
                         deviations if `std` was supplied) must be the same
                         length as the inputs! {len(ci)} != {len(split_ids)}"""
                    )

                ids_to_uncertainties = {
                    split_ids[i]: {"ci_lower": p[0], "ci_upper": p[1], "std": s}
                    for i, (p, s) in enumerate(zip(ci, std))
                }
                self.results[fold_key][self._UNCERTAINTY_KEY] = ids_to_uncertainties
            else:
                self.results[fold_key][self._UNCERTAINTY_KEY] = None

            if not isinstance(params, (dict, type(None))):
                raise TypeError(
                    f"Parameters must be stored as a dictionary, not {type(params)}!"
                )
            params = immutify_dictionary(params) if params else params
            self.results[fold_key][self._PARAMS_KEY] = params if params else {}
            self.is_recorded[fold_number] = True

            logger.info(
                f"Recorded fold " f"{self.dataset_name}-{fold_number} successfully."
            )

            truth = self._get_data_from_df(split_ids, as_type="tuple")[1]
            self.results[fold_key][self._SCORES_KEY] = score_array(
                truth, predictions, self.metadata.task_type
            )
            logger.debug(
                f"Scored fold '" f"{self.dataset_name}-{fold_key} successfully."
            )

    def as_dict(self):
        """Return a MatbenchTask object as a dictionary.

        Required method from MSONAble.

        Returns:
            (dict)
        """
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            self._BENCHMARK_KEY: self.benchmark_name,
            self._DATASET_KEY: self.dataset_name,
            self._RESULTS_KEY: dict(self.results),
        }

    def validate(self):
        """Validate a task after all folds have been recorded.

        There are a few requirements for a task to be validated:
        - Data types of each predicted sample must match those
            specified by the validation procedure
        - All folds must be recorded
        - There must be no extra or missing required keys from
            the data, including indices. Every index specified in
            the validation procedure must be present in its
            correct fold, and no extras may be present.
        - Ensure consistency of the supplied uncertainty values.
            For example, if std is specified and ci is specified
            for one sample, it must be specified for all samples.
            If ci is specified but std is not, that must be
            consistent for all samples.
        Returns:

        """
        self._check_all_folds_recorded(
            f"Cannot validate task {self.dataset_name} "
            f"unless all folds recorded!"
        )
        task_type = self.metadata.task_type

        # Check for extra fold keys
        extra_fold_keys = [k for k in self.results if k not in self.folds_keys]
        if extra_fold_keys:
            raise KeyError(
                f"Extra fold keys {extra_fold_keys} for task "
                f"{self.dataset_name} not allowed."
            )

        for fold_key in self.folds_keys:
            if fold_key not in self.results:
                raise KeyError(
                    f"Required fold data for fold '{fold_key}' "
                    f"for task {self.dataset_name} not found."
                )

            # Check for extra or missing keys inside each fold:
            # need params, scores, and data.
            req_subfold_keys = [self._SCORES_KEY, self._DATA_KEY, self._PARAMS_KEY]
            extra_subfold_keys = [
                k for k in self.results[fold_key] if k not in req_subfold_keys
            ]
            if self._UNCERTAINTY_KEY in extra_subfold_keys:
                extra_subfold_keys.remove(self._UNCERTAINTY_KEY)
            if extra_subfold_keys:
                raise KeyError(
                    f"Extra keys {extra_subfold_keys} for fold results of "
                    f"'{fold_key}' for task {self.dataset_name}  not allowed."
                )
            req_subfold_keys.append(self._UNCERTAINTY_KEY)
            for subkey in req_subfold_keys:
                fold_results = self.results[fold_key]
                if (
                    subkey is not self._UNCERTAINTY_KEY
                    and subkey not in fold_results
                ):
                    raise KeyError(
                        f"Required key '{subkey}' for task {self.dataset_name} "
                        f"not found for fold '{fold_key}'."
                    )
                if subkey == self._SCORES_KEY:
                    scores = self.results[fold_key][subkey]
                    metrics = REG_METRICS if task_type == REG_KEY else CLF_METRICS
                    for m in metrics:
                        if m not in scores:
                            raise KeyError(
                                f"Required score '{m}' for task "
                                f"{self.dataset_name} "
                                f"not found for '{fold_key}'."
                            )
                        elif not isinstance(scores[m], float):
                            raise TypeError(
                                f"Required score '{m}' for task "
                                f"{self.dataset_name} "
                                f"is not float-type for '{fold_key}'!"
                            )
                    extra_metrics = [k for k in scores if k not in metrics]
                    if extra_metrics:
                        raise KeyError(
                            f"Extra keys {extra_metrics} for fold scores of "
                            f"'{fold_key}' for task {self.dataset_name} "
                            f"not allowed."
                        )

                # results data indices are cast by json to be strings,
                # so must be converted to int
                elif subkey == self._DATA_KEY:
                    fold_data = self.results[fold_key].data

                    # Ensure all the indices are present with no
                    # extras for each fold
                    req_indices = set(self.validation[fold_key].test)
                    remaining_indices = copy.deepcopy(req_indices)
                    extra_indices = {}
                    if self.metadata.task_type == REG_KEY:
                        allowed_types = (float,)
                    else:
                        allowed_types = (bool, float)

                    for ix, datum in fold_data.items():
                        if ix not in req_indices:
                            extra_indices[ix] = datum
                        else:
                            if not isinstance(datum, allowed_types):
                                raise TypeError(
                                    f"Data point '{ix}: {datum}' has data type "
                                    f"{type(datum)} while required type is "
                                    f"{allowed_types} for task "
                                    f"{self.dataset_name} !"
                                )
                            if self.metadata.task_type == CLF_KEY:
                                if isinstance(datum, float):
                                    if datum < 0 or datum > 1:
                                        raise ValueError(
                                            f"Probability estimate '{ix}': {datum}"
                                            f"for task {self.dataset_name} outside "
                                            f"of range [0, 1]."
                                        )

                            remaining_indices.remove(ix)

                    if extra_indices and not remaining_indices:
                        raise ValueError(
                            f"{len(extra_indices)} extra indices for problem "
                            f"{self.dataset_name} are not allowed (found in "
                            f"{fold_key}: {remaining_indices}"
                        )
                    elif not extra_indices and remaining_indices:
                        raise ValueError(
                            f"{len(remaining_indices)} required indices "
                            f"for problem {self.dataset_name} not "
                            f"found for {fold_key}: {remaining_indices}"
                        )
                    elif extra_indices and remaining_indices:
                        raise ValueError(
                            f"{len(remaining_indices)} required indices "
                            f"for problem {self.dataset_name} not "
                            f"found and {len(extra_indices)} not "
                            f"allowed indices found for {fold_key}!"
                        )
                    else:
                        pass

                elif subkey == self._UNCERTAINTY_KEY:
                    if self._UNCERTAINTY_KEY in self.results[fold_key].keys():
                        uncertainties = self.results[fold_key][subkey]
                    else:
                        uncertainties = None
                    if uncertainties is not None:
                        std = uncertainties["std"]
                        ci = uncertainties["ci"]

                        if all(isinstance(s, float) for s in std):
                            if any(isinstance(c, float) for c in ci):
                                if not all(isinstance(c, float) for c in ci):
                                    raise ValueError(
                                        "std specified for all samples "
                                        "but ci not specified for some."
                                    )
                        else:
                            if any(isinstance(s, float) for s in std):
                                raise ValueError(
                                    "std is specified for some, but not for all."
                                )

                        if all(isinstance(c, float) for c in ci):
                            if any(isinstance(s, float) for s in std):
                                if not all(isinstance(s, float) for s in std):
                                    raise ValueError(
                                        "ci specified for all samples "
                                        "but ci not specified for some."
                                    )
                        else:
                            if any(isinstance(c, float) for c in ci):
                                raise ValueError(
                                    "ci is specified for some, but not for all."
                                )

                # Params key has no required form;
                # it is up to the model to determine it.

        logger.debug(f"Data for {self.dataset_name} successfully validated.")

    @property
    def scores(self):
        """Comprehensive score metrics for this task.

        Gets means, maxes, mins, and more distribution stats (across folds)
        for all scoring metrics defined for this task.

        There will be different scores for classification problems and
        regression problems.

        Returns:
            (dict): A dictionary of all the scores for this task.
        """
        metric_keys = (
            REG_METRICS if self.metadata.task_type == REG_KEY else CLF_METRICS
        )
        scores = {}
        self._check_all_folds_recorded("Cannot score unless all folds are recorded!")
        for mk in metric_keys:
            metric = {}

            # scores for a metric among all folds
            raw_metrics_on_folds = [
                self.results[fk][self._SCORES_KEY][mk]
                for fk in self.folds_map.values()
            ]
            for op in FOLD_DIST_METRICS:
                metric[op] = getattr(np, op)(raw_metrics_on_folds)
            scores[mk] = metric
        return RecursiveDotDict(scores)

    @property
    def is_recorded(self):
        """Determine what folds in the task are recorded.

        Returns:
            ({int: bool}): Keys are fold numbers, values are whether the
                fold is recorded or not.
        """
        is_recorded = {}
        for fnum, fkey in self.folds_map.items():
            if self.results[fkey][self._DATA_KEY]:
                is_recorded[fnum] = True
            else:
                is_recorded[fnum] = False
        return is_recorded

    @property
    def all_folds_recorded(self):
        """Determine if all folds are recorded.

        Returns:
            (bool): True if all folds are recorded, False otherwise.
        """
        return all([v for v in self.is_recorded.values()])

    @property
    def has_polymorphs(self):
        """Determine if a task's raw data contains polymorphs.

        Returns:
            (bool) If true, contains polymorphs.
        """
        checker_key = "pmg_composition"
        self._check_is_loaded()
        if self.metadata.input_type == "composition":
            stc = StrToComposition(target_col_id=checker_key, reduce=True)
            comps = stc.featurize_dataframe(self.df, "composition")[
                checker_key
            ].values
        elif self.metadata.input_type == "structure":
            stc = StructureToComposition(target_col_id=checker_key, reduce=True)
            comps = stc.featurize_dataframe(self.df, "structure")[checker_key].values
        else:
            raise ValueError(
                "Cannot check for polymorphs without input type in "
                "(structure, composition)!"
            )

        unique_comps = set(comps)
        if len(unique_comps) != len(comps):
            return True
        else:
            return False



import unittest

from matbench.constants import CLF_KEY, REG_KEY
from matbench.metadata import mbv01_metadata, mbv01_validation


class TestMetadata(unittest.TestCase):
    def test_mbv01_metadata(self):

        # for matbench v0.1
        for ds, metadata in mbv01_metadata.items():
            for key in ["task_type", "n_samples", "input_type", "target"]:
                self.assertIn(key, metadata.keys())

                if key == "input_type":
                    self.assertIn(metadata[key], ["composition", "structure"])

                elif key == "task_type":
                    self.assertIn(metadata[key], [REG_KEY, CLF_KEY])
        self.assertEqual(len(list(mbv01_metadata.values())), 13)

        self.assertIn("metadata", mbv01_validation)
        self.assertIn("splits", mbv01_validation)
        self.assertEqual(len(mbv01_validation.splits), 13)

        for k in mbv01_metadata:
            self.assertEqual(len(mbv01_validation.splits[k]), 5)



import os
import glob
import unittest

from monty.serialization import loadfn

from matbench.bench import MatbenchBenchmark


BENCHMARKS_DIR = os.path.join(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir), "benchmarks")

LOCAL = False
# LOCAL = True

INFO_FILE = "info.json"
RESULTS_FILE = "results.json.gz"
SRC_NB_GLOB = "*.ipynb"
SRC_PY_GLOB = "*.py"


class BenchmarkSubmissionTest(unittest.TestCase):

    def setUp(self) -> None:
        print(f"Running test in benchmark dir: {BENCHMARKS_DIR}")
        print(f"Running test locally? {LOCAL}")

    def test_dir_structure_correct(self):
        self.assertTrue(os.path.exists(BENCHMARKS_DIR))

        # make sure the dir only has submissions as directories
        for file_or_dir in os.listdir(BENCHMARKS_DIR):
            full_path = os.path.join(BENCHMARKS_DIR, file_or_dir)
            if not full_path.endswith(".DS_Store"):
                self.assertTrue(os.path.isdir(full_path))

            if not LOCAL:
                for banned_file in (".ipynb_checkpoints",):
                    self.assertFalse(os.path.exists(os.path.join(full_path, banned_file)))

    def test_submissions(self):

        algo_names = []
        for bmark_d in os.listdir(BENCHMARKS_DIR):
            bmark_path = os.path.join(BENCHMARKS_DIR, bmark_d)
            if os.path.isdir(bmark_path):
                print(f"checking benchmark {bmark_path}")
                for required_file in (INFO_FILE, RESULTS_FILE):

                    full_path = os.path.join(bmark_path, required_file)
                    self.assertTrue(os.path.exists(full_path))

                    if required_file == INFO_FILE:
                        info = loadfn(full_path)
                        self.assertIn("notes", info)

                        for field_must_not_be_empty in ("authors", "algorithm", "algorithm_long", "bibtex_refs", "requirements"):
                            self.assertIsNotNone(info.get(field_must_not_be_empty, None))

                        algo_names.append(info["algorithm"])

                    if required_file == RESULTS_FILE:
                        mb = MatbenchBenchmark.from_file(full_path)
                        is_valid = mb.is_valid
                        self.assertTrue(is_valid)

                        print(f"{RESULTS_FILE} file in {bmark_path}: valid ({is_valid}), complete ({mb.is_complete}), recorded ({mb.is_recorded})")


                # benchmark must either contain a notebook or at least one .py file
                nb_paths = glob.glob(os.path.join(bmark_path, SRC_NB_GLOB))
                py_paths = glob.glob(os.path.join(bmark_path, SRC_PY_GLOB))
                self.assertGreater(len(nb_paths + py_paths), 0)

                # ensure there are no large (10MB+) files in the submission
                # apart from the results
                for file in os.listdir(bmark_path):
                    if file != RESULTS_FILE:
                        fpath = os.path.join(bmark_path, file)
                        fsize_mb = os.stat(fpath).st_size/1e6
                        self.assertLessEqual(fsize_mb, 10)

            else:
                print(f"path {bmark_path} skipped!")

        # Each algo name must be unique
        self.assertEqual(len(algo_names), len(set(algo_names)))


if __name__ == "__main__":
    unittest.main()


import os
import requests
import json


from matbench import __version__



def release_gh():
    payload = {
        "tag_name": "v" + __version__,
        "target_commitish": "main",
        "name": "v" + __version__,
        "body": "",
        "draft": False,
        "prerelease": False
    }
    response = requests.post(
        "https://api.github.com/repos/hackingmaterials/matbench/releases",
        data=json.dumps(payload),
        headers={
            "Authorization": "token " + os.environ["GITHUB_RELEASES_TOKEN"]})
    print(response.text)


if __name__ == "__main__":
    release_gh()



import pandas as pd
from monty.serialization import dumpfn
from sklearn.model_selection import KFold, StratifiedKFold

from matbench.constants import (
    CLF_KEY,
    REG_KEY,
    TEST_KEY,
    TRAIN_KEY,
    VALIDATION_METADATA_KEY,
    VALIDATION_SPLIT_KEY,
)
from matbench.data_ops import load
from matbench.metadata import mbv01_metadata

pd.set_option("display.max_rows", 500)
pd.set_option("display.max_columns", 500)
pd.set_option("display.width", 1000)


def matbench_v01():
    """
    Generate validation json file for matbench v0.1, a benchmark
    of 13 regression and binary classification datasets which
    uses nested cross validation (random splits for regression,
    stratified for classification).

    The end result is a dictionary of the form:

    {
        "metadata": arbitrary dict metadata about this split scheme,
        "splits": {
            dataset_name_1: {
                "fold_0": {
                    "train": [matbench_id 1, matbench_id 2, ...]
                    "test": [matbench_id 3783, matbench_id 4874, ...]
                    },
                ...
                "fold_N": ...
            },
            dataset_name_2: { ... },
            ...
    }

    The "splits" key adheres to this specific format. The "metadata"
    key is freeform. Both "splits" and "metadata" must be present
    in the validation file generated.


    Returns:
        None
    """
    d = {
        VALIDATION_METADATA_KEY: {
            "n_splits": 5,
            "random_state": 18012019,
            "shuffle": True,
        },
        VALIDATION_SPLIT_KEY: None,
    }

    kfold_config = d[VALIDATION_METADATA_KEY]

    splits = {}

    for ds, info in mbv01_metadata.items():
        split = {}
        task_type = info.task_type

        if task_type == REG_KEY:
            kfold = KFold(**kfold_config)
        elif task_type == CLF_KEY:
            kfold = StratifiedKFold(**kfold_config)
        else:
            raise ValueError(
                f"'problem_type' must be one of {[REG_KEY, CLF_KEY]}, not '{task_type}'."
            )

        df = load(ds)
        print(df)

        # y is needed for stratified splits
        fold = 0
        for train_ix, test_ix in kfold.split(X=df, y=df[mbv01_metadata[ds].target]):

            train_df = df.iloc[train_ix]
            test_df = df.iloc[test_ix]

            train_mbid = train_df.index.tolist()
            test_mbid = test_df.index.tolist()

            split[f"fold_{fold}"] = {TRAIN_KEY: train_mbid, TEST_KEY: test_mbid}
            fold += 1

        splits[ds] = split

    d["splits"] = splits

    print("Writing file...")
    dumpfn(d, "../matbench/matbench_v0.1_validation.json")


if __name__ == "__main__":
    matbench_v01()



import os
from setuptools import setup, find_packages

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(MODULE_DIR, "requirements.txt"), "r") as f:
    requirements = f.read().replace(" ", "").split("\n")

# source of version is in the constants file
VERSION_FILE = os.path.join(MODULE_DIR, "matbench/constants.py")
token = "VERSION = "
with open(VERSION_FILE, "r") as f:
    version = None
    for line in f.readlines():
        if token in line:
            version = line.replace(token, "").strip()
# Double quotes are contained in the read line, remove them
version = version.replace("\"", "")


if __name__ == "__main__":
    setup(
        name='matbench',
        version=version,
        description='a machine learning benchmark for materials science',
        long_description="A machine learning benchmark for materials science. "
                         "https://github.com/materialsproject/matbench",
        url='https://github.com/materialsproject/matbench',
        author=['Alex Dunn', 'Anubhav Jain'],
        author_email='ardunn@lbl.gov',
        license='modified BSD',
        packages=find_packages(where="."),
        package_data={
            "matbench": ["*.json"],
            "matbench.tests": ["*.json"]
        },
        zip_safe=False,
        install_requires=requirements,
        extras_require={},
        test_suite='matbench',
        tests_require='tests',
        include_package_data=True
    )



"""
Metadata for matbench.
"""

from matminer.datasets.utils import _load_dataset_dict
from monty.serialization import loadfn

from matbench.constants import (
    MBV01_DATASET_METADATA_PATH,
    MBV01_VALIDATION_DATA_PATH,
)
from matbench.util import RecursiveDotDict

MATMINER_DATASET_METADATA = _load_dataset_dict()

mbv01_metadata = loadfn(MBV01_DATASET_METADATA_PATH)
for d in mbv01_metadata.keys():
    mbv01_metadata[d].update(MATMINER_DATASET_METADATA[d])
mbv01_metadata = RecursiveDotDict(mbv01_metadata)


mbv01_validation = loadfn(MBV01_VALIDATION_DATA_PATH)
mbv01_validation = RecursiveDotDict(mbv01_validation)



import copy
import json
import os
import unittest

from matbench.bench import MatbenchBenchmark
from matbench.constants import (
    CLF_KEY,
    COMPOSITION_KEY,
    MBV01_KEY,
    REG_KEY,
    STRUCTURE_KEY,
)
from matbench.task import MatbenchTask
from matbench.tests.util import FULL_TEST, TEST_DIR, model_random


class TestMatbenchBenchmark(unittest.TestCase):
    def setUp(self) -> None:
        self.msonability_tmp_path = os.path.join(
            TEST_DIR, "msonability_test_augmented.json"
        )
        self.static_full_bench_json = os.path.join(
            TEST_DIR, "mb_all_tasks_random.json"
        )
        self.static_3_bench_json = os.path.join(TEST_DIR, "mb_3_tasks_random.json")

    def tearDown(self) -> None:
        for f in [self.msonability_tmp_path]:
            if os.path.exists(f):
                os.remove(f)

    def test_from_preset(self):
        for preset, n_tasks in {
            REG_KEY: 10,
            CLF_KEY: 3,
            STRUCTURE_KEY: 9,
            COMPOSITION_KEY: 4,
            "all": 13,
        }.items():
            mb = MatbenchBenchmark.from_preset(
                benchmark=MBV01_KEY, preset_name=preset, autoload=False
            )
            self.assertEqual(len(mb.tasks_map.keys()), n_tasks)

    def test_scores(self):
        mb = MatbenchBenchmark.from_file(self.static_full_bench_json)
        scores = mb.scores
        self.assertAlmostEqual(
            14.169387576348942, scores.matbench_dielectric.mape.mean, places=10
        )
        self.assertAlmostEqual(
            2.376419875235826, scores.matbench_mp_e_form.rmse.mean, places=10
        )
        self.assertAlmostEqual(
            0.49560975609756097, scores.matbench_expt_is_metal.f1.min, places=10
        )

    def test_info(self):
        mb = MatbenchBenchmark.from_file(self.static_full_bench_json)
        mb.info
        mb.get_info()

    def test_add_metadata(self):
        mb = MatbenchBenchmark.from_file(self.static_full_bench_json)
        example_metadata = {"some": "metadata"}
        mb.add_metadata(example_metadata)
        self.assertDictEqual(example_metadata, mb.user_metadata)

        bad_metadata = [1, 2, 3]
        mb.add_metadata(bad_metadata)
        self.assertDictEqual(example_metadata, mb.user_metadata)

    def test_complete_valid_recorded(self):
        mb = MatbenchBenchmark.from_file(self.static_full_bench_json)
        self.assertTrue(mb.is_valid)
        self.assertTrue(mb.is_recorded)
        self.assertTrue(mb.is_complete)

        # if one sample is missing from one task fold, throw error
        mb.matbench_dielectric.results.fold_0.data.pop("mb-dielectric-0010")
        self.assertFalse(mb.is_valid)
        self.assertTrue(len(mb.validate().keys()), 1)

        # if one task is missing one fold
        mb = MatbenchBenchmark.from_file(self.static_full_bench_json)
        mb.matbench_mp_e_form.results.pop("fold_1")
        self.assertFalse(mb.is_valid)
        self.assertTrue(len(mb.validate().keys()), 1)

        # if one fold is not recorded
        mb = MatbenchBenchmark.from_file(self.static_full_bench_json)
        mb.matbench_phonons.results.fold_1 = {}
        self.assertFalse(mb.is_recorded)

        # if one or more tasks is missing
        mb = MatbenchBenchmark.from_file(self.static_3_bench_json)
        self.assertFalse(mb.is_complete)

    def test_MSONability(self):
        # Test serialization
        self.generate_benchmark_json_files(write=False, full_set=FULL_TEST)

        # Test deserialization
        mb = MatbenchBenchmark.from_file(self.static_full_bench_json)
        self.assertTrue(mb.is_complete)
        self.assertTrue(mb.is_valid)
        self.assertTrue(mb.is_recorded)

        # Ensure object can be added to and re/de-serialized without issue
        example_metadata = {"some": "metadata"}
        mb.add_metadata(example_metadata)
        mb.to_file(self.msonability_tmp_path)
        mb = MatbenchBenchmark.from_file(self.msonability_tmp_path)
        self.assertTrue(mb.is_complete)
        self.assertTrue(mb.is_valid)
        self.assertTrue(mb.is_recorded)
        self.assertDictEqual(example_metadata, mb.user_metadata)

        with open(self.static_full_bench_json, "r") as f:
            d_truth = json.load(f)

        # Test if hash is wrong (hash error should occur)
        d_test = copy.deepcopy(d_truth)
        d_test[MatbenchBenchmark._HASH_KEY] = "not a real hash"
        with self.assertRaises(ValueError):
            MatbenchBenchmark.from_dict(d_test)

        # Test if datetime is wrong (hash error should occur)
        d_test = copy.deepcopy(d_truth)
        d_test[MatbenchBenchmark._DATESTAMP_KEY] = "2021.02.02 06:33.21"
        with self.assertRaises(ValueError):
            MatbenchBenchmark.from_dict(d_test)

        # Test if one or more task names is wrong
        d_test = copy.deepcopy(d_truth)
        d_test[MatbenchBenchmark._TASKS_KEY]["matbench_dielectric"][
            MatbenchTask._DATASET_KEY
        ] = "matbench_expt_gap"
        d_test[MatbenchBenchmark._TASKS_KEY]["matbench_expt_gap"][
            MatbenchTask._DATASET_KEY
        ] = "matbench_dielectric"
        with self.assertRaises(ValueError):
            MatbenchBenchmark.from_dict(d_test)

        # Test if one or more benchmark names is wrong
        d_test = copy.deepcopy(d_truth)
        d_test[MatbenchBenchmark._TASKS_KEY]["matbench_dielectric"][
            MatbenchTask._BENCHMARK_KEY
        ] = "matbench_v0.2"
        with self.assertRaises(ValueError):
            MatbenchBenchmark.from_dict(d_test)

    def test_completeness(self):
        mb = MatbenchBenchmark(benchmark=MBV01_KEY, subset=["matbench_steels"])

        self.assertFalse(mb.is_complete)
        self.assertFalse(mb.is_composition_complete)
        self.assertFalse(mb.is_structure_complete)

        mb = MatbenchBenchmark.from_preset(MBV01_KEY, preset_name=COMPOSITION_KEY)
        self.assertFalse(mb.is_complete)
        self.assertTrue(mb.is_composition_complete)
        self.assertFalse(mb.is_structure_complete)

        mb = MatbenchBenchmark.from_preset(MBV01_KEY, preset_name=STRUCTURE_KEY)
        self.assertFalse(mb.is_complete)
        self.assertFalse(mb.is_composition_complete)
        self.assertTrue(mb.is_structure_complete)

        mb = MatbenchBenchmark.from_preset(MBV01_KEY, preset_name="all")
        self.assertTrue(mb.is_complete)
        self.assertTrue(mb.is_composition_complete)
        self.assertTrue(mb.is_structure_complete)

        mb = MatbenchBenchmark.from_preset(MBV01_KEY, preset_name=REG_KEY)
        self.assertFalse(mb.is_complete)
        self.assertTrue(mb.is_regression_complete)
        self.assertFalse(mb.is_classification_complete)

        mb = MatbenchBenchmark.from_preset(MBV01_KEY, preset_name=CLF_KEY)
        self.assertFalse(mb.is_complete)
        self.assertFalse(mb.is_regression_complete)
        self.assertTrue(mb.is_classification_complete)

    def generate_benchmark_json_files(self, write=True, full_set=False):
        """
        Generate new benchmark files from a random model.

        Args:
            write:
            full_set:

        Returns:

        """

        if full_set:
            subset = None  # use all tasks
        else:
            subset = [
                "matbench_dielectric",
                "matbench_steels",
                "matbench_glass",
            ]  # to generate 3 task file

        mb = MatbenchBenchmark(benchmark=MBV01_KEY, autoload=False, subset=subset)

        for task in mb.tasks:
            task.load()
            for fold in task.folds:
                train_inputs, train_outputs = task.get_train_and_val_data(fold)
                test_inputs = task.get_test_data(fold, include_target=False)
                model_response = model_random(
                    train_outputs, test_inputs, response_type=task.metadata.task_type
                )
                task.record(
                    fold,
                    model_response,
                    params={"some_param": 1, "other_param": 12.39348},
                )

        if subset:
            if write:
                mb.to_file(self.static_3_bench_json)
            self.assertTrue(mb.is_recorded)
            self.assertTrue(mb.is_valid)
            self.assertFalse(mb.is_complete)
        else:
            if write:
                mb.to_file(self.static_full_bench_json)
            self.assertTrue(mb.is_recorded)
            self.assertTrue(mb.is_valid)
            self.assertTrue(mb.is_complete)



import logging
import os
import unittest

import numpy as np

from matbench.tests.util import TEST_DIR
from matbench.util import (
    MSONable2File,
    RecursiveDotDict,
    hash_dictionary,
    immutify_dictionary,
    initialize_logger,
)


class TestUtils(unittest.TestCase):
    def setUp(self) -> None:
        self.json_file = os.path.join(TEST_DIR, "example.json")
        self.log_file = os.path.join(TEST_DIR, "matbench_test.log")

    def tearDown(self) -> None:
        if os.path.exists(self.json_file):
            os.remove(self.json_file)

        if os.path.exists(self.log_file):
            os.remove(self.log_file)

    def test_RecursiveDotDict(self):

        d = {"a": {"b": [1, 2, 3], "c": "string"}, "b": 12}

        rd = RecursiveDotDict(d)

        self.assertDictEqual(rd.a, {"b": [1, 2, 3], "c": "string"})
        self.assertListEqual(rd.a.b, [1, 2, 3])
        self.assertEqual(rd.b, 12)

    def test_MSONable2File(self):
        class MSONable2FileChild(MSONable2File):
            def __init__(self, prop1):
                self.prop1 = prop1

            def as_dict(self):
                return {"prop1": self.prop1}

            @classmethod
            def from_dict(cls, d):
                return cls(prop1=d["prop1"])

        ms = MSONable2FileChild(prop1="blue")
        ms.to_file(self.json_file)

        ms2 = MSONable2FileChild.from_file(self.json_file)
        self.assertEqual(ms2.prop1, "blue")

    def test_logging(self):
        logger = initialize_logger(
            "matbench_test", log_dir=TEST_DIR, level=logging.DEBUG
        )

        logger.info("example msg 1")
        logger.debug("Example msg 2")
        logger.critical("Example msg 3")

        self.assertTrue(os.path.exists(self.log_file))

        with open(self.log_file, "r") as f:
            logtxt = f.read()
            self.assertIn("INFO", logtxt)
            self.assertIn("DEBUG", logtxt)
            self.assertIn("CRITICAL", logtxt)
            self.assertEqual(len(logtxt.split("\n")), 4)


class TestHashingDictionaryFunctions(unittest.TestCase):
    def setUp(self) -> None:
        self.d = {
            "q": [1, 2, 3],
            "b": {
                "c": "12",
                "d": 15,
                "e": np.asarray([4, 5, 6]),
                "f": {"z": 12, "a": [7, 8]},
            },
            "c": np.int64(400),
        }
        self.d_same = {
            "q": [1, 2, 3],
            "b": {
                "c": "12",
                "d": 15,
                "e": np.asarray([4, 5, 6]),
                "f": {"a": [7, 8], "z": 12},
            },
            "c": 400,
        }
        self.d_different = {
            "q": [1, 2, 3],
            "b": {
                "c": "12",
                "d": 15,
                "e": np.asarray([4, 5, 6]),
                "f": {"z": 12, "a": [7, 9]},
            },
            "c": 400,
        }

    def test_immutify_dictionary(self):
        d_immutable = immutify_dictionary(self.d)
        d_truth = {
            "q": (1, 2, 3),
            "b": {"c": "12", "d": 15, "e": (4, 5, 6), "f": {"z": 12, "a": (7, 8)}},
            "c": 400,
        }
        self.assertDictEqual(d_immutable, d_truth)

    def test_hash_dictionary(self):
        self.assertEqual(hash_dictionary(self.d), hash_dictionary(self.d_same))
        self.assertNotEqual(
            hash_dictionary(self.d), hash_dictionary(self.d_different)
        )
